import numpy as np

from ase.lattice.cubic import BodyCenteredCubic
from ase.constraints import FixAtoms, StrainFilter
from ase.optimize import FIRE
from ase.build import bulk
from ase.calculators.lammpslib import LAMMPSlib
from ase.units import GPa  # unit conversion
from ase.lattice.cubic import SimpleCubicFactory
from ase.io import read

from matscipy.neighbours import neighbour_list, mic
from matscipy.elasticity import fit_elastic_constants

def make_screw_cyl(alat, C11, C12, C44,
                   cylinder_r=10, cutoff=5.5,
                   hard_core=False,
                   center=[0., 0., 0.],
                   l_extend=[0., 0., 0.],
                   symbol='W'):

    """Makes screw dislocation using atomman library

    Parameters
    ----------
    alat : float
        Lattice constant of the material.
    C11 : float
        C11 elastic constant of the material.
    C12 : float
        C12 elastic constant of the material.
    C44 : float
        C44 elastic constant of the material.
    cylinder_r : float
        radius of cylinder of unconstrained atoms around the
        dislocation  in angstrom
    cutoff : float
        Potential cutoff for marinica potentials for FS cutoff = 4.4
    hard_core : bool
        Description of parameter `hard_core`.
    center : type
        The position of the dislocation core and the center of the
                 cylinder with FixAtoms condition
    l_extend : float
        extension of the box. used for creation of initial
        dislocation position with box equivalent to the final position
    symbol : string
        Symbol of the element to pass to ase.lattice.cubic.SimpleCubicFactory
        default is "W" for tungsten

    Returns
    -------
    disloc : ase.Atoms object
        screw dislocation cylinder.
    bulk : ase.Atoms object
        bulk disk used to generate dislocation
    u : np.array
        displacement per atom.
    """
    # https://github.com/usnistgov/atomman
    from atomman import ElasticConstants
    from atomman.defect import Stroh

    # Create a Stroh object with junk data
    stroh = Stroh(ElasticConstants(C11=141, C12=110, C44=98),
                  np.array([0, 0, 1]))

    axes = np.array([[1, 1, -2],
                     [-1, 1, 0],
                     [1, 1, 1]])

    c = ElasticConstants(C11=C11, C12=C12, C44=C44)
    burgers = alat * np.array([1., 1., 1.])/2.

    # Solving a new problem with Stroh.solve
    stroh.solve(c, burgers, axes=axes)

    # test the solution that it does not crash
    # pos_test = uc.set_in_units(np.array([12.4, 13.5, -10.6]), 'angstrom')
    # disp = stroh.displacement(pos_test)
    # print("displacement =", uc.get_in_units(disp, 'angstrom'), 'angstrom')

    unit_cell = BodyCenteredCubic(directions=axes.tolist(),
                                  size=(1, 1, 1), symbol=symbol,
                                  pbc=(False, False, True),
                                  latticeconstant=alat)

    # make the dislocation core center of the box
    disloCenterX = alat * np.sqrt(6.)/6.0
    disloCenterY = alat * np.sqrt(2.)/6.0

    unit_cell.positions[:, 0] -= disloCenterX
    unit_cell.positions[:, 1] -= disloCenterY

    # shift to move the fixed atoms boundary condition for the
    # configuration with shifted dislocation core
    shift_x = 2.0 * center[0]
    shift_y = 2.0 * center[1]

    l_shift_x = 2.0 * l_extend[0]
    l_shift_y = 2.0 * l_extend[1]

    # size of the cubic cell as a 112 direction
    Lx = int(round((cylinder_r + 3.*cutoff + shift_x + l_shift_x)
                   / (alat * np.sqrt(6.))))

    # size of the cubic cell as a 110 direction
    Ly = int(round((cylinder_r + 3.*cutoff + shift_y + l_shift_y)
                   / (alat * np.sqrt(2.))))
    # factor 2 to make sure odd number of images is translated
    # it is important for the correct centering of the dislocation core
    bulk = unit_cell * (2*Lx, 2*Ly, 1)
    # make 0, 0, at the center
    bulk.positions[:, 0] -= Lx * alat * np.sqrt(6.)
    bulk.positions[:, 1] -= Ly * alat * np.sqrt(2.)

    # wrap
    # bulk.set_scaled_positions(bulk.get_scaled_positions())
    # apply shear here:
    # bulk.cell *= D
    # bulk.positions *= D

    x, y, z = bulk.positions.T

    radius_x_y_zero = np.sqrt(x**2 + y**2)
    mask_zero = radius_x_y_zero < cylinder_r + 2.*cutoff

    radius_x_y_center = np.sqrt((x - center[0])**2 + (y - center[1])**2)
    mask_center = radius_x_y_center < cylinder_r + 2.*cutoff

    radius_x_y_l_shift = np.sqrt((x - l_extend[0])**2 + (y - l_extend[1])**2)
    mask_l_shift = radius_x_y_l_shift < cylinder_r + 2.*cutoff

    final_mask = mask_center | mask_zero | mask_l_shift
    # leave only atoms inside the cylinder
    bulk = bulk[final_mask]

    disloc = bulk.copy()
    # calculate and apply the displacements for atomic positions
    u = stroh.displacement(bulk.positions - center)
    u = -u if hard_core else u

    disloc.positions += u
    x, y, z = disloc.positions.T

    radius_x_y_zero = np.sqrt(x**2 + y**2)
    mask_zero = radius_x_y_zero > cylinder_r

    radius_x_y_center = np.sqrt((x - center[0])**2 + (y - center[1])**2)
    mask_center = radius_x_y_center > cylinder_r

    radius_x_y_l_shift = np.sqrt((x - l_extend[0])**2 + (y - l_extend[1])**2)
    mask_l_shift = radius_x_y_l_shift > cylinder_r

    fix_mask = mask_center & mask_zero & mask_l_shift
    # leave only atoms inside the cylinder
    fix_atoms = FixAtoms(mask=fix_mask)
    disloc.set_constraint(fix_atoms)

    # make an "region" array to map bulk and fixed atoms
    # all atoms are "MM" by default
    region = np.full_like(disloc, "MM")
    region[fix_mask] = np.full_like(disloc[fix_mask], "fixed")
    disloc.new_array("region", region)

    # center the atoms to avoid "lost atoms" error by lammps
    center_shift = np.diagonal(bulk.cell).copy()
    center_shift[2] = 0.0  # do not shift along z direction

    disloc.positions += center_shift / 2.0
    bulk.positions += center_shift / 2.0

    return disloc, bulk, u


def make_edge_cyl(alat, C11, C12, C44,
                  cylinder_r=10, cutoff=5.5,
                  symbol='W'):
    '''
    makes edge dislocation using atomman library

    cylinder_r - radius of cylinder of unconstrained atoms around the
                 dislocation  in angstrom

    cutoff - potential cutoff for marinica potentials for FS cutoff = 4.4

    symbol : string
        Symbol of the element to pass to ase.lattuce.cubic.SimpleCubicFactory
        default is "W" for tungsten
    '''
    # https://github.com/usnistgov/atomman
    from atomman import ElasticConstants
    from atomman.defect import Stroh
    # Create a Stroh ojbect with junk data
    stroh = Stroh(ElasticConstants(C11=141, C12=110, C44=98),
                                   np.array([0, 0, 1]))

    axes = np.array([[1, 1, 1],
                     [1, -1, 0],
                     [1, 1, -2]])

    c = ElasticConstants(C11=C11, C12=C12, C44=C44)
    burgers = alat * np.array([1., 1., 1.])/2.

    # Solving a new problem with Stroh.solve
    # Does not work with the new version of atomman
    stroh.solve(c, burgers, axes=axes)

    unit_cell = BodyCenteredCubic(directions=axes.tolist(),
                                  size=(1, 1, 1), symbol='W',
                                  pbc=(False, False, True),
                                  latticeconstant=alat)

    bulk = unit_cell.copy()

    # shift to make the zeros of the cell betweem the atomic planes
    # and under the midplane on Y axes
    X_midplane_shift = (1.0/3.0)*alat*np.sqrt(3.0)/2.0
    Y_midplane_shift = 0.25*alat*np.sqrt(2.0)

    bulk_shift = [X_midplane_shift,
                  Y_midplane_shift,
                  0.0]

    bulk.positions += bulk_shift

    tot_r = cylinder_r + cutoff + 0.01

    Lx = int(round(tot_r/(alat*np.sqrt(3.0)/2.0)))
    Ly = int(round(tot_r/(alat*np.sqrt(2.))))

    # factor 2 to make shure odd number of images is translated
    # it is important for the correct centering of the dislocation core
    bulk = bulk * (2*Lx, 2*Ly, 1)

    center_shift = [Lx * alat * np.sqrt(3.0)/2.,
                    Ly * alat * np.sqrt(2.),
                    0.0]

    bulk.positions -= center_shift

    ED = bulk.copy()

    disp = stroh.displacement(ED.positions)

    ED.positions += disp

    x, y, z = ED.positions.T
    radius_x_y_zero = np.sqrt(x**2 + y**2)
    mask = radius_x_y_zero < tot_r

    ED = ED[mask]
    bulk = bulk[mask]

    bulk.write("before.xyz")

    ED.write("after_disp.xyz")

    x, y, z = ED.positions.T
    radius_x_y_zero = np.sqrt(x**2 + y**2)
    mask_zero = radius_x_y_zero > cylinder_r
    fix_atoms = FixAtoms(mask=mask_zero)

    ED.set_constraint(fix_atoms)

    x, y, z = bulk.positions.T
    # move lower left segment
    bulk.positions[(y < 0.0) & (x < X_midplane_shift)] -= \
        [alat * np.sqrt(3.0) / 2.0, 0.0, 0.0]
    # make the doslocation extraplane center
    bulk.positions += [(1.0/3.0)*alat*np.sqrt(3.0)/2.0, 0.0, 0.0]

    return ED, bulk


def plot_vitek(dislo, bulk,
               alat=3.16, plot_axes=None, xyscale=10):
    """Plots vitek map from ase configurations.

    Parameters
    ----------
    dislo : ase.Atoms
        Dislocation configuration.
    bulk : ase.Atoms
        Corresponding bulk configuration for calculation of displacements.
    alat : float
        Lattice parameter for calculation of neghbour list cutoff.
    plot_axes : matplotlib.Axes.axes object
        Existing axes to plot on, allows to pass existing matplotlib axes
        have full control of the graph outside the function.
        Makes possible to plot multiple differential displacement
        maps using subplots.
        Default is None, then new graph is created by plt.subplots()
        Description of parameter `plot_axes`.
    xyscale : float
        xyscale of the graph

    Returns
    -------
    None

    """
    from atomman import load
    from atomman.defect import differential_displacement


    lengthB = 0.5*np.sqrt(3.)*alat
    burgers = np.array([0.0, 0.0, lengthB])

    base_system = load("ase_Atoms", bulk)
    disl_system = load("ase_Atoms", dislo)

    neighborListCutoff = 0.95 * alat

    # plot window is +-10 angstroms from center in x,y directions, and one Burgers vector
    # thickness along z direction
    x, y, _ = bulk.positions.T

    plot_range = np.array([[x.mean() - xyscale, x.mean() + xyscale],
                          [y.mean() - xyscale, y.mean() + xyscale],
                          [-0.1, alat * 3.**(0.5) / 2.]])

    # This scales arrows such that b/2 corresponds to the
    # distance between atoms on the plot
    plot_scale = 1.885618083

    differential_displacement(base_system, disl_system,
                              burgers,
                              cutoff=neighborListCutoff,
                              xlim=plot_range[0],
                              ylim=plot_range[1],
                              zlim=plot_range[2],
                              plot_scale=plot_scale,
                              plot_axes=plot_axes)

    return None


def show_NEB_configurations(images, bulk, xyscale=7,
                            show=True, core_positions=None):
    """Plots Vitek differential displacement maps for the list of images
       for example along the NEB path.

    Parameters
    ----------
    images : list of ase.Atoms
        List of configurations with dislocations.
    bulk : ase.Atoms
        Corresponding bulk configuration for calculation of displacements.
    xyscale : float
        xyscale of the graph
    show : bool
        Show the figure after plotting. Default is True.
    core_positions : list
        [x, y] position of dislocation core to plot

    Returns
    -------
    figure
        If the show is False else returns None

    """
    import matplotlib.pyplot as plt

    n_images = len(images)
    fig2 = plt.figure(figsize=(n_images * 4, 4))

    for i, image in enumerate(images):
        ax1 = fig2.add_subplot("1%i%i" % (n_images, i+1))
        plot_vitek(image, bulk, plot_axes=ax1, xyscale=xyscale)
        if core_positions is not None:
            x, y = core_positions[i]
            ax1.scatter(x, y, marker="+", s=200, c='C1')
    if show:
        fig2.show()
        return None
    else:
        return fig2


def show_configuration(disloc, bulk, u, fixed_mask=None):
    """shows the displacement fixed atoms."""
    import matplotlib.pyplot as plt
    fig = plt.figure(figsize=(16, 4))

    ax1 = fig.add_subplot(131)
    ax1.set_title("z displacement, $\AA$")
    sc = ax1.scatter(bulk.positions[:, 0], bulk.positions[:, 1], c=u.T[2])
    ax1.axvline(0.0, color="red", linestyle="dashed")
    ax1.set_xlabel("x, $\AA$")
    ax1.set_ylabel("y, $\AA$")
    plt.colorbar(sc)

    ax2 = fig.add_subplot(132)
    ax2.set_title("x displacement, $\AA$")
    sc = ax2.scatter(bulk.positions[:, 0], bulk.positions[:, 1], c=u.T[0])
    ax2.set_xlabel("x, $\AA$")
    ax2.set_ylabel("y, $\AA$")
    plt.colorbar(sc, format="%.1e")

    ax3 = fig.add_subplot(133)
    ax3.set_title("y displacement, $\AA$")
    sc = ax3.scatter(bulk.positions[:, 0], bulk.positions[:, 1], c=u.T[1])
    plt.colorbar(sc, format="%.1e")
    ax3.set_xlabel("x, $\AA$")
    ax3.set_ylabel("y, $\AA$")

    if fixed_mask is not None:

        ax1.scatter(bulk.positions[fixed_mask, 0],
                    bulk.positions[fixed_mask, 1], c="k")

        ax2.scatter(bulk.positions[fixed_mask, 0],
                    bulk.positions[fixed_mask, 1], c="k")

        ax3.scatter(bulk.positions[fixed_mask, 0],
                    bulk.positions[fixed_mask, 1], c="k")

    plt.show()

    return None


def get_elastic_constants(pot_path=None,
                          calculator=None,
                          delta=1e-2,
                          symbol="W"):
    """
    return lattice parameter, and cubic elastic constants: C11, C12, 44
    using matscipy function
    pot_path - path to the potential

    symbol : string
        Symbol of the element to pass to ase.lattuce.cubic.SimpleCubicFactory
        default is "W" for tungsten
    """

    unit_cell = bulk(symbol)

    if (pot_path is not None) and (calculator is None):
        # create lammps calculator with the potential
        lammps = LAMMPSlib(lmpcmds=["pair_style eam/fs",
                           "pair_coeff * * %s W" % pot_path],
                           atom_types={'W': 1}, keep_alive=True)
        calculator = lammps

    unit_cell.set_calculator(calculator)

#   simple calculation to get the lattice constant and cohesive energy
#    alat0 = W.cell[0][1] - W.cell[0][0]
    sf = StrainFilter(unit_cell)  # or UnitCellFilter(W) -> to minimise wrt pos, cell
    opt = FIRE(sf)
    opt.run(fmax=1e-4)  # max force in eV/A
    alat = unit_cell.cell[0][1] - unit_cell.cell[0][0]
#    print("a0 relaxation %.4f --> %.4f" % (a0, a))
#    e_coh = W.get_potential_energy()
#    print("Cohesive energy %.4f" % e_coh)

    Cij, Cij_err = fit_elastic_constants(unit_cell,
                                         symmetry="cubic",
                                         delta=delta)

    Cij = Cij/GPa  # unit conversion to GPa

    elasticMatrix3x3 = Cij[:3, :3]
    # average of diagonal elements: C11, C22, C33
    C11 = elasticMatrix3x3.diagonal().mean()
    # make mask to extract non diagonal elements
    mask = np.ones((3, 3), dtype=bool)
    np.fill_diagonal(mask, False)

    # average of all non diagonal elements from 1 to 3
    C12 = elasticMatrix3x3[mask].mean()

    # average of diagonal elements from 4 till 6: C44, C55, C66,
    C44 = Cij[3:, 3:].diagonal().mean()

    # A = 2.*C44/(C11 - C12)

    if (pot_path is not None) and (calculator is None):
        lammps.lmp.close()

    return alat, C11, C12, C44


def make_barrier_configurations(elastic_param=None,
                                pot_path=None, calculator=None,
                                cylinder_r=10, hard_core=False, **kwargs):
    """Creates the initial and final configurations for the NEB calculation
        The positions in FixedAtoms constrained region are average between
        final and initial configurations

    Parameters
    ----------
    pot_path : string
        Path to the potential file.
    calculator : type
        Description of parameter `calculator`.
    cylinder_r : float
        Radius of cylinder of unconstrained atoms around the
                    dislocation  in angstrom.
    hard_core : bool
        Type of the core hard or soft.
        If hard is chosen the displacement field is reversed.
    **kwargs :
        Keyword arguments to pass to make_screw_cyl() function.

    Returns
    -------
    disloc_ini : ase.Atoms
        Initial dislocation configuration.
    disloc_fin : ase.Atoms
        Final dislocation configuration.
    bulk : ase.Atoms
        Perfect bulk configuration for Vitek displacement maps

    """

    if pot_path is not None:
        alat, C11, C12, C44 = get_elastic_constants(pot_path=pot_path)
        # get the cutoff from the potential file
        with open(pot_path) as potfile:
            for i, tmp_str in enumerate(potfile):
                if i == 4:  # read the last number in the fifth line
                    cutoff = float(tmp_str.split()[-1])
                    break

    elif calculator is not None:
        alat, C11, C12, C44 = get_elastic_constants(calculator=calculator)
        cutoff = 5.0  # the value for trainig data for GAP from paper

    elif elastic_param is not None:
        alat, C11, C12, C44 = elastic_param
        cutoff = 5.5

    cent_x = np.sqrt(6.0)*alat/3.0
    center = [cent_x, 0.0, 0.0]

    disloc_ini, bulk_ini, __ = make_screw_cyl(alat, C11, C12, C44,
                                              cylinder_r=cylinder_r,
                                              cutoff=cutoff,
                                              hard_core=hard_core,
                                              l_extend=center, **kwargs)

    disloc_fin, __, __ = make_screw_cyl(alat, C11, C12, C44,
                                        cylinder_r=cylinder_r,
                                        cutoff=cutoff,
                                        hard_core=hard_core,
                                        center=center, **kwargs)

    # get the fixed atoms constrain
    FixAtoms = disloc_ini.constraints[0]
    # get the indices of fixed atoms
    fixed_atoms_indices = FixAtoms.get_indices()

    # make the average position of fixed atoms
    # between initial and the lastlast position
    ini_fix_pos = disloc_ini.get_positions()[fixed_atoms_indices]
    fin_fix_pos = disloc_fin.get_positions()[fixed_atoms_indices]

    new_av_pos = (ini_fix_pos + fin_fix_pos)/2.0

    positions = disloc_ini.get_positions()
    positions[fixed_atoms_indices] = new_av_pos
    disloc_ini.set_positions(positions, apply_constraint=False)

    positions = disloc_fin.get_positions()
    positions[fixed_atoms_indices] = new_av_pos
    disloc_fin.set_positions(positions, apply_constraint=False)

    return disloc_ini, disloc_fin, bulk_ini

def make_screw_cyl_kink(alat, C11, C12, C44,
                        cylinder_r=40, kink_length=26, kind="double", **kwargs):
    """Function to create kink configuration based on make_screw_cyl() function.
        Double kink configuration is in agreement with quadrupoles in terms of formation energy.
        Single kink configurations provide correct and stable structure, but formation energy is not accessible?

    Parameters
    ----------
    alat : float
        Lattice constant of the material.
    C11 : float
        C11 elastic constant of the material.
    C12 : float
        C12 elastic constant of the material.
    C44 : float
        C44 elastic constant of the material.
    cylinder_r : float
        radius of cylinder of unconstrained atoms around the
        dislocation  in angstrom
    kink_length : int
        Length of the cell per kink along b in unit of b, must be even.
    kind : string
        kind of the kink: right, left or double
    **kwargs :
        Keyword arguments to pass to make_screw_cyl() function.

    Returns
    -------
    kink : ase.atoms
        kink configuration
    reference_straight_disloc : ase.atoms
        reference straight dislocation configuration
    large_bulk : ase.atoms
        large bulk cell corresponding to the kink configuration
    """
    b = np.sqrt(3.0) * alat / 2.0
    cent_x = np.sqrt(6.0) * alat / 3.0

    disloc_ini, disloc_fin, bulk_ini = make_barrier_configurations((alat, C11, C12, C44),
                                                                    cylinder_r=cylinder_r, **kwargs)

    if kind == "double":

        large_bulk = bulk_ini * [1, 1, 2 * kink_length]
        reference_straight_disloc = disloc_ini * [1, 1, 2 * kink_length]

        if kink_length % 2:
            print("WARNING: length is not even!")

        kink = disloc_ini * [1, 1, kink_length // 2]
        middle_kink = disloc_fin * [1, 1, kink_length]

        middle_kink.positions += np.array((0.0, 0.0, kink.get_cell()[2][2]))

        kink.constraints[0].index = np.append(kink.constraints[0].index,
                                              middle_kink.constraints[0].get_indices() + len(kink))
        kink.extend(middle_kink)
        kink.cell[2][2] += middle_kink.cell[2][2]

        upper_kink = disloc_ini * [1, 1, kink_length // 2]
        upper_kink.positions += np.array((0.0, 0.0, kink.get_cell()[2][2]))

        kink.constraints[0].index = np.append(kink.constraints[0].index,
                                              upper_kink.constraints[0].get_indices() + len(kink))
        kink.extend(upper_kink)
        kink.cell[2][2] += upper_kink.cell[2][2]

    elif kind == "right":

        large_bulk = bulk_ini * [1, 1, kink_length]
        reference_straight_disloc = disloc_ini * [1, 1, kink_length]

        kink = disloc_ini * [1, 1, kink_length // 2]
        upper_disloc = disloc_fin * [1, 1, kink_length // 2]

        upper_disloc.positions += np.array((0.0, 0.0, kink.cell[2][2]))

        kink.extend(upper_disloc)
        kink.constraints[0].index = np.append(kink.constraints[0].index,
                                              upper_disloc.constraints[0].get_indices() + len(kink))
        kink.cell[2][2] += upper_disloc.cell[2][2]

        # we have to adjust the cell to make the kink vector periodic
        # here we remove two atomic rows. it is nicely explained in the paper
        _, _, z = large_bulk.positions.T
        right_kink_mask = z < large_bulk.cell[2][2] - 2.0 * b / 3 - 0.01

        kink = kink[right_kink_mask]

        cell = kink.cell.copy()

        # right kink is created when the kink vector is in positive x direction
        # assuming (x, y, z) is right group of vectors
        cell[2][0] += cent_x
        cell[2][2] -= 2.0 * b / 3.0
        kink.set_cell(cell, scale_atoms=False)

        # make sure all the atoms are removed and cell is modified for the bulk as well.
        large_bulk.cell[2][0] += cent_x
        large_bulk.cell[2][2] -= 2.0 * b / 3.0
        large_bulk = large_bulk[right_kink_mask]
        for constraint in kink.constraints:
            large_bulk.set_constraint(constraint)

    elif kind == "left":

        large_bulk = bulk_ini * [1, 1, kink_length]
        reference_straight_disloc = disloc_ini * [1, 1, kink_length]

        kink = disloc_fin * [1, 1, kink_length // 2]

        upper_disloc = disloc_ini * [1, 1, kink_length // 2]
        upper_disloc.positions += np.array((0.0, 0.0, kink.cell[2][2]))

        kink.extend(upper_disloc)
        kink.constraints[0].index = np.append(kink.constraints[0].index,
                                              upper_disloc.constraints[0].get_indices() + len(kink))
        kink.cell[2][2] += upper_disloc.cell[2][2]

        # we have to adjust the cell to make the kink vector periodic
        # here we remove one atomic row. it is nicely explained in the paper
        _, _, z = large_bulk.positions.T
        left_kink_mask = z < large_bulk.cell[2][2] - 1.0 * b / 3 - 0.01

        kink = kink[left_kink_mask]

        cell = kink.cell.copy()

        # left kink is created when the kink vector is in negative x direction
        # assuming (x, y, z) is right group of vectors
        cell[2][0] -= cent_x
        cell[2][2] -= 1.0 * b / 3.0
        kink.set_cell(cell, scale_atoms=False)

        # make sure all the atoms are removed and cell is modified for the bulk as well.
        large_bulk.cell[2][0] -= cent_x
        large_bulk.cell[2][2] -= 1.0 * b / 3.0
        large_bulk = large_bulk[left_kink_mask]
        for constraint in kink.constraints:
            large_bulk.set_constraint(constraint)

    else:
        raise ValueError('Kind must be "right", "left" or "double"')

    return kink, reference_straight_disloc, large_bulk

def slice_long_dislo(kink, kink_bulk, b):
    """Function to slice a long dislocation configuration to perform dislocation structure and core position analysis

    Parameters
    ----------
    kink : ase.Atoms
        kink configuration to slice
    kink_bulk : ase.Atoms
        corresponding bulk configuration to perform mapping for slicing
    b : float
        burgers vector b should be along z direction


    Returns
    -------
    sliced_kink : list of [sliced_bulk, sliced_kink]
        sliced configurations 1 b length each
    disloc_z_positions : float
        positions of each sliced configuration (center along z)
    """

    if not len(kink) == len(kink_bulk):
        raise ValueError('"kink" and "kink_bulk" must be same size')

    n_slices = int(np.round(kink.cell[2][2] / b * 3))
    atom_z_positions = kink_bulk.positions.T[2]

    kink_z_length = kink_bulk.cell[2][2]

    sliced_kink = []
    disloc_z_positions = []

    for slice_id in range(n_slices):

        shift = slice_id * b / 3.0

        upper_bound = 5.0 * b / 6.0 + shift
        lower_bound = -b / 6.0 + shift

        if upper_bound < kink_z_length:

            mask = np.logical_and(atom_z_positions < upper_bound,
                                  atom_z_positions > lower_bound)

            bulk_slice = kink_bulk.copy()[mask]
            kink_slice = kink.copy()[mask]

        else:  # take into account PBC at the end of the box

            upper_mask = atom_z_positions < (upper_bound - kink_z_length)

            mask = np.logical_or(upper_mask,
                                 atom_z_positions > lower_bound)

            bulk_slice = kink_bulk.copy()[mask]
            kink_slice = kink.copy()[mask]

            # move the bottom atoms on top of the box
            kink_slice.positions[upper_mask[mask]] += np.array(kink_bulk.cell[2])

            bulk_slice.positions[upper_mask[mask]] += np.array((kink_bulk.cell[2]))

        # print(kink_bulk[mask].positions.T[2].max())
        # print(kink_bulk[mask].positions.T[2].min())

        bulk_slice.positions -= np.array((0.0, 0.0, shift))
        kink_slice.positions -= np.array((0.0, 0.0, shift))

        bulk_slice.cell = kink_bulk.cell
        bulk_slice.cell[2][2] = b
        bulk_slice.cell[2][0] = 0

        kink_slice.cell = kink_bulk.cell
        kink_slice.cell[2][2] = b
        kink_slice.cell[2][0] = 0

        sliced_kink.append([bulk_slice, kink_slice])
        disloc_z_positions.append(b / 3.0 + shift)

    disloc_z_positions = np.array(disloc_z_positions)

    return sliced_kink, disloc_z_positions


def compare_configurations(dislo, bulk, dislo_ref, bulk_ref,
                           alat, cylinder_r=None, print_info=True):
    """Compares two dislocation configurations based on the gradient of
       the displacements along the bonds.

    Parameters
    ----------
    dislo : ase.Atoms
        Dislocation configuration.
    bulk : ase.Atoms
        Corresponding bulk configuration for calculation of displacements.
    dislo_ref : ase.Atoms
        Reference dislocation configuration.
    bulk_ref : ase.Atoms
        Corresponding reference bulk configuration
        for calculation of displacements.
    alat : float
        Lattice parameter for calculation of neghbour list cutoff.
    cylinder_r : float or None
        Radius of region of comparison around the dislocation coreself.
        If None makes global comparison based on the radius of
        `dislo` configuration, else compares the regions with `cylinder_r`
        around the dislocation core position.
    print_info : bool
        Flag to switch print statement about the type of the comparison
    Returns
    -------
    float
        The Du norm of the differences per atom.

    """

    if cylinder_r is None:
        x, y, __ = bulk.positions.T
        radius = np.sqrt(x**2 + y**2)
        cutoff_radius = radius.max() - 10.

        if print_info:
            print("Making a global comparison with radius %.2f" % cutoff_radius)

    else:
        cutoff_radius = cylinder_r
        if print_info:
            print("Making a local comparison with radius %.2f" % cutoff_radius)

    x, y, __ = bulk_ref.positions.T
    radius = np.sqrt(x**2 + y**2)
    cutoff_mask = (radius < cutoff_radius)

    second_NN_distance = alat
    bulk_i, bulk_j = neighbour_list('ij', bulk_ref, second_NN_distance)

    I_core, J_core = np.array([(i, j) for i, j in zip(bulk_i, bulk_j) if cutoff_mask[i]]).T

    mapping = {}

    for i in range(len(bulk)):
        mapping[i] = np.linalg.norm(bulk_ref.positions -
                                    bulk.positions[i], axis=1).argmin()

    u_ref = dislo_ref.positions - bulk_ref.positions

    u = dislo.positions - bulk.positions
    u_extended = np.zeros(u_ref.shape)
    u_extended[list(mapping.values()), :] = u

    du = u_extended - u_ref

    Du = np.linalg.norm(np.linalg.norm(mic(du[J_core, :] - du[I_core, :],
                                           bulk.cell), axis=1))
    return Du


def cost_function(pos, dislo, bulk, cylinder_r, elastic_param,
                  hard_core=False, print_info=True):
    """Cost function for fitting analytical displacement field
       and detecting dislocation core position. Uses `compare_configurations`
       function for the minimisation of the core position.

    Parameters
    ----------
    pos : list of float
        Positions of the core to build the analytical solution [x, y].
    dislo : ase.Atoms
        Dislocation configuration.
    bulk : ase.Atoms
        Corresponding bulk configuration for calculation of displacements.
    cylinder_r : float or None
        Radius of region of comparison around the dislocation coreself.
        If None makes global comparison based on the radius of
        `dislo` configuration, else compares the regions with `cylinder_r`
        around the dislocation core position.
    elastic_param : list of float
        List containing alat, C11, C12, C44
    hard_core : bool
        type of the core True for hard
    print_info : bool
        Flag to switch print statement about the type of the comparison

    Returns
    -------
    float
        Error for optimisation (result from `compare_configurations` function)

    """
    # https://github.com/usnistgov/atomman
    from atomman import ElasticConstants
    from atomman.defect import Stroh

    # Create a Stroh ojbect with junk data
    stroh = Stroh(ElasticConstants(C11=141, C12=110, C44=98),
                  np.array([0, 0, 1]))

    axes = np.array([[1, 1, -2],
                    [-1, 1, 0],
                    [1, 1, 1]])

    alat, C11, C12, C44 = elastic_param

    c = ElasticConstants(C11=C11, C12=C12, C44=C44)
    burgers = alat * np.array([1., 1., 1.])/2.

    # Solving a new problem with Stroh.solve
    stroh.solve(c, burgers, axes=axes)

    center = (pos[0], pos[1], 0.0)
    u = stroh.displacement(bulk.positions - center)
    u = -u if hard_core else u

    dislo_guess = bulk.copy()
    dislo_guess.positions += u

    err = compare_configurations(dislo, bulk,
                                 dislo_guess, bulk,
                                 alat, cylinder_r=cylinder_r,
                                 print_info=print_info)

    return err


def screw_cyl_tetrahedral(alat, C11, C12, C44,
                          scan_r=15,
                          symbol="W",
                          imp_symbol='H',
                          hard_core=False,
                          center=(0., 0., 0.)):
    """Generates a set of terahedral positions with `scan_r` radius.
       Applies the screw dislocation displacement for creating an initial guess
       for the H positions at dislocation core.

    Parameters
    ----------
    alat : float
        Lattice constant of the material.
    C11 : float
        C11 elastic constant of the material.
    C12 : float
        C12 elastic constant of the material.
    C44 : float
        C44 elastic constant of the material.
    scan_r : float
        Radius of the region to create tetrahedral positions.
    symbol : string
        Symbol of the element to pass to ase.lattuce.cubic.SimpleCubicFactory
        default is "W" for tungsten
    imp_symbol : string
        Symbol of the elemnt to pass creat Atoms object
        default is "H" for hydrogen
    hard_core : float
        Type of the dislocatino core if True then -u
        (sign of displacement is flipped) is applied.
        Default is False i.e. soft core is created.


    center : tuple of floats
        Coordinates of dislocation core (center) (x, y, z).
        Default is (0., 0., 0.)

    Returns
    -------
    ase.Atoms object
        Atoms object with predicted tetrahedral
        positions around dislocation core.

    """
    # https://github.com/usnistgov/atomman
    from atomman import ElasticConstants
    from atomman.defect import Stroh

    axes = np.array([[1, 1, -2],
                     [-1, 1, 0],
                     [1, 1, 1]])

    unit_cell = BodyCenteredCubic(directions=axes.tolist(),
                                  size=(1, 1, 1), symbol=symbol,
                                  pbc=(False, False, True),
                                  latticeconstant=alat)

    BCCTetras = BodyCenteredCubicTetrahedralFactory()

    impurities = BCCTetras(directions=axes.tolist(),
                           size=(1, 1, 1),
                           symbol=imp_symbol,
                           pbc=(False, False, True),
                           latticeconstant=alat)

    impurities = impurities[impurities.positions.T[2] < alat*1.2]
    impurities.set_cell(unit_cell.get_cell())
    impurities.wrap()

    disloCenterY = alat * np.sqrt(2.)/6.0
    disloCenterX = alat * np.sqrt(6.)/6.0

    impurities.positions[:, 0] -= disloCenterX
    impurities.positions[:, 1] -= disloCenterY

    # size of the cubic cell as a 112 direction
    Lx = int(round((scan_r)/(alat * np.sqrt(6.))))

    # size of the cubic cell as a 110 direction
    Ly = int(round((scan_r) / (alat * np.sqrt(2.))))
    # factor 2 to ,ake shure odd number of images is translated
    # it is important for the correct centering of the dislocation core
    bulk_tetra = impurities * (2*(Lx + 1), 2*(Ly + 1), 1)
    # make 0, 0, at the center

    # make 0, 0, at the center
    bulk_tetra.positions[:, 0] -= (Lx + 1) * alat * np.sqrt(6.) - center[0]
    bulk_tetra.positions[:, 1] -= (Ly + 1) * alat * np.sqrt(2.) - center[1]

    x, y, z = bulk_tetra.positions.T

    radius_x_y_zero = np.sqrt(x**2 + y**2)
    mask_zero = radius_x_y_zero < scan_r

    radius_x_y_center = np.sqrt((x - center[0])**2 + (y - center[1])**2)
    mask_center = radius_x_y_center < scan_r

    final_mask = mask_center | mask_zero
    # leave only atoms inside the cylinder
    bulk_tetra = bulk_tetra[final_mask]

    # Create a Stroh ojbect with junk data
    stroh = Stroh(ElasticConstants(C11=141, C12=110, C44=98),
                  np.array([0, 0, 1]))

    c = ElasticConstants(C11=C11, C12=C12, C44=C44)
    burgers = alat * np.array([1., 1., 1.])/2.

    # Solving a new problem with Stroh.solve
    stroh.solve(c, burgers, axes=axes)

    dislo_tetra = bulk_tetra.copy()

    impurities_u = stroh.displacement(bulk_tetra.positions - center)
    impurities_u = -impurities_u if hard_core else impurities_u

    dislo_tetra.positions += impurities_u

    return dislo_tetra


def screw_cyl_octahedral(alat, C11, C12, C44,
                         scan_r=15,
                         symbol="W",
                         imp_symbol='H',
                         hard_core=False,
                         center=(0., 0., 0.)):
    """Generates a set of octahedral positions with `scan_r` radius.
       Applies the screw dislocation displacement for creating an initial guess
       for the H positions at dislocation core.

    Parameters
    ----------
    alat : float
        Lattice constant of the material.
    C11 : float
        C11 elastic constant of the material.
    C12 : float
        C12 elastic constant of the material.
    C44 : float
        C44 elastic constant of the material.
    symbol : string
        Symbol of the element to pass to ase.lattuce.cubic.SimpleCubicFactory
        default is "W" for tungsten
    imp_symbol : string
        Symbol of the elemnt to pass creat Atoms object
        default is "H" for hydrogen
    symbol : string
        Symbol of the elemnt to pass creat Atoms object
    hard_core : float
        Type of the dislocatino core if True then -u
        (sign of displacement is flipped) is applied.
        Default is False i.e. soft core is created.
    center : tuple of floats
        Coordinates of dislocation core (center) (x, y, z).
        Default is (0., 0., 0.)

    Returns
    -------
    ase.Atoms object
        Atoms object with predicted tetrahedral
        positions around dislocation core.

    """
    # TODO: Make one function for impurities and pass factory to it:
    # TODO: i.e. octahedral or terahedral

    # https://github.com/usnistgov/atomman
    from atomman import ElasticConstants
    from atomman.defect import Stroh

    axes = np.array([[1, 1, -2],
                     [-1, 1, 0],
                     [1, 1, 1]])

    unit_cell = BodyCenteredCubic(directions=axes.tolist(),
                                  size=(1, 1, 1), symbol=symbol,
                                  pbc=(False, False, True),
                                  latticeconstant=alat)

    BCCOctas = BodyCenteredCubicOctahedralFactory()

    impurities = BCCOctas(directions=axes.tolist(),
                          size=(1, 1, 1), symbol=imp_symbol,
                          pbc=(False, False, True),
                          latticeconstant=alat)

    impurities = impurities[impurities.positions.T[2] < alat*1.2]
    impurities.set_cell(unit_cell.get_cell())
    impurities.wrap()

    disloCenterY = alat * np.sqrt(2.)/6.0
    disloCenterX = alat * np.sqrt(6.)/6.0

    impurities.positions[:, 0] -= disloCenterX
    impurities.positions[:, 1] -= disloCenterY

    L = int(round(2.0*scan_r/(alat*np.sqrt(2.)))) + 1

    bulk_octa = impurities * (L, L, 1)

    # make 0, 0, at the center
    bulk_octa.positions[:, 0] -= L * alat * np.sqrt(6.)/2. - center[0]
    bulk_octa.positions[:, 1] -= L * alat * np.sqrt(2.)/2. - center[1]

    x, y, z = bulk_octa.positions.T

    radius_x_y_zero = np.sqrt(x**2 + y**2)
    mask_zero = radius_x_y_zero < scan_r

    radius_x_y_center = np.sqrt((x - center[0])**2 + (y - center[1])**2)
    mask_center = radius_x_y_center < scan_r

    final_mask = mask_center | mask_zero
    # leave only atoms inside the cylinder
    bulk_octa = bulk_octa[final_mask]

    # Create a Stroh object with junk data
    stroh = Stroh(ElasticConstants(C11=141, C12=110, C44=98),
                                   np.array([0, 0, 1]))

    c = ElasticConstants(C11=C11, C12=C12, C44=C44)
    burgers = alat * np.array([1., 1., 1.])/2.

    # Solving a new problem with Stroh.solve
    stroh.solve(c, burgers, axes=axes)

    dislo_octa = bulk_octa.copy()

    impurities_u = stroh.displacement(bulk_octa.positions - center)
    impurities_u = -impurities_u if hard_core else impurities_u

    dislo_octa.positions += impurities_u

    return dislo_octa


class BodyCenteredCubicTetrahedralFactory(SimpleCubicFactory):
    """A factory for creating tetrahedral lattices in bcc structure"""
    xtal_name = "bcc_tetrahedral"

    bravais_basis = [[0.0, 0.5, 0.25],
                     [0.0, 0.5, 0.75],
                     [0.0, 0.25, 0.5],
                     [0.0, 0.75, 0.5],
                     [0.5, 0.0, 0.75],
                     [0.25, 0.0, 0.5],
                     [0.75, 0.0, 0.5],
                     [0.5, 0.0, 0.25],
                     [0.5, 0.25, 0.0],
                     [0.5, 0.75, 0.0],
                     [0.25, 0.5, 0.0],
                     [0.75, 0.5, 0.0]]


class BodyCenteredCubicOctahedralFactory(SimpleCubicFactory):
    """A factory for creating octahedral lattices in bcc structure"""
    xtal_name = "bcc_octahedral"

    bravais_basis = [[0.0, 0.0, 0.5],
                     [0.5, 0.0, 0.0],
                     [0.5, 0.0, 0.5],
                     [0.0, 0.5, 0.0],
                     [0.0, 0.5, 0.5],
                     [0.5, 0.5, 0.0]]


def dipole_displacement_angle(W_bulk, dislo_coord_left, dislo_coord_right,
                              shift=0.0, mode=1.0):
    """
        Generates a simple displacement field for two dislocations in a dipole
        configuration uding simple Voltera solution as u = b/2 * angle
    """


    burgers = W_bulk.cell[2][2]

    shifted_positions = W_bulk.positions + shift - dislo_coord_left
    x, y, __ = shifted_positions.T
    displacement_left = np.arctan2(y, x) * burgers / (2.0 * np.pi)

    shifted_positions = W_bulk.positions + shift - dislo_coord_right
    x, y, __ = shifted_positions.T
    displacement_right = np.arctan2(y, mode*x) * burgers / (2.0 * np.pi)

    # make two easy core configurations

    u_dipole = np.zeros_like(W_bulk.positions)
    u_dipole.T[2] = displacement_left - mode*displacement_right

    return u_dipole


def get_u_img(W_bulk,
              dislo_coord_left,
              dislo_coord_right,
              n_img=10, n1_shift=0, n2_shift=0):
    """
        Function for getting displacemnt filed for images of quadrupole cells
        used by `make_screw_quadrupole`
    """

    u_img = np.zeros_like(W_bulk.positions)

    C1_quadrupole, C2_quadrupole, __ = W_bulk.get_cell()

    for n1 in range(-n_img, n_img + 1):
        for n2 in range(-n_img, n_img + 1):

            shift = n1 * C1_quadrupole + n2 * C2_quadrupole

            if n1 != n1_shift or n2 != n2_shift:

                u_img += dipole_displacement_angle(W_bulk,
                                                   dislo_coord_left + shift,
                                                   dislo_coord_right + shift,
                                                   shift=n2_shift*C2_quadrupole + n1_shift*C1_quadrupole)

    return u_img


def make_screw_quadrupole(alat,
                          left_shift=0,
                          right_shift=0,
                          n1u=5,
                          symbol="W"):
    """Generates a screw dislocation dipole configuration
       for effective quadrupole arrangement. Works for BCC systems.

    Parameters
    ----------
    alat : float
        Lattice parameter of the system in Angstrom.
    left_shift : float, optional
        Shift of the left dislocation core in number of dsitances to next
        equivalent disocation core positions needed for creation for final
        configuration for NEB. Default is 0.
    right_shift : float, optional
        shift of the right dislocation core in number of dsitances to next
        equivalent disocation core positions needed for creation for final
        configuration for NEB. Default is 0.
    n1u : int, odd number
        odd number! length of the cell a doubled distance between core along x.
        Main parameter to calculate cell vectors
    symbol : string
        Symbol of the element to pass to ase.lattuce.cubic.SimpleCubicFactory
        default is "W" for tungsten

    Returns
    -------
    disloc_quadrupole : ase.Atoms
        Resulting quadrupole configuration.
    W_bulk : ase.Atoms
        Perfect system.
    dislo_coord_left : list of float
        Coodrinates of left dislocation core [x, y]
    dislo_coord_right : list of float
        Coodrinates of right dislocation core [x, y]

    Notes
    -----

    Calculation of cell vectors
    +++++++++++++++++++++++++++
    From [1]_ we take:

    - Unit vectors for the cell are:

    .. math::

        u = \frac{1}{3}[1 \bar{2} 1];

    .. math::

        v = \frac{1}{3}[2 \bar{1} \bar{1}];

    .. math::

        z = b = \frac{1}{2}[1 1 1];

    - Cell vectors are:

    .. math::

        C_1 = n^u_1 u + n^v_1 v + C^z_1 z;

    .. math::

        C_2 = n^u_2 u + n^v_2 v + C^z_2 z;

    .. math::

        C_3 = z

    - For quadrupole arrangement n1u needs to be odd number,
      for 135 atoms cell we take n1u=5

    - To have quadrupole as as close as possible to a square one has to take:
    .. math::

        2 n^u_2 + n^v_2 = n^u_1

    .. math::

        n^v_2 \approx \frac{n^u_1}{\sqrt{3}}

    - for n1u = 5:

    .. math::

        n^v_2 \approx \frac{n^u_1}{\sqrt{3}} = 2.89 \approx 3.0

    .. math::

        n^u_2 = \frac{1}{2} (n^u_1 - n^v_2) \approx \frac{1}{2} (5-3)=1

    - Following [2]_ cell geometry is optimized by ading tilt compomemts
    Cz1 and Cz2 for our case of n1u = 3n - 1:

    Easy core

    .. math::

        C^z_1 = + \frac{1}{3}

    .. math::

        C^z_2 = + \frac{1}{6}

    Hard core

    .. math::

        C^z_1 = + \frac{1}{3}

    .. math::

        C^z_2 = + \frac{1}{6}

    may be typo in the paper check the original!

    References:
    +++++++++++

    .. [1] Ventelon, L. & Willaime, F. J 'Core structure and Peierls potential
       of screw dislocations in alpha-Fe from first principles: cluster versus
       dipole approaches' Computer-Aided Mater Des (2007) 14(Suppl 1): 85.
       https://doi.org/10.1007/s10820-007-9064-y

    .. [2] Cai W. (2005) Modeling Dislocations Using a Periodic Cell.
       In: Yip S. (eds) Handbook of Materials Modeling. Springer, Dordrecht
       https://link.springer.com/chapter/10.1007/978-1-4020-3286-8_42

    """

    unit_cell = BodyCenteredCubic(directions=[[1, -2, 1],
                                      [2, -1, -1],
                                      [1, 1, 1]],
                                  symbol=symbol,
                                  pbc=(True, True, True),
                                  latticeconstant=alat, debug=0)

    unit_cell_u, unit_cell_v, unit_cell_z = unit_cell.get_cell()

    # calculate the cell vectors according to the Ventelon paper
    # the real configrution depends on rounding check it here

    n2v = int(np.rint(n1u/np.sqrt(3.0)))
    # when the n1u - n2v difference is odd it is impossible to have
    # perfect arrangemt of translation along x with C2 equal to 0.5*n1u
    # choice of rounding between np.ceil() and np.trunc() makes a different
    # configuration but of same quality of arrangement of quadrupoles (test)
    n2u = np.ceil((n1u - n2v) / 2.)
    n1v = 0

    print("Not rounded values of C2 componets: ")
    print("n2u: %.2f,  n2v: %.2f" % ((n1u - n2v) / 2., n1u/np.sqrt(3.0)))
    print("Calculated cell vectors from n1u = %i" % n1u)
    print("n1v = %i" % n1v)
    print("n2u = %i" % n2u)
    print("n2v = %i" % n2v)

    bulk = unit_cell.copy()*[n1u, n2v, 1]

    # add another periodic shift in x direction to c2 vector
    # for proper periodicity (n^u_2=1) of the effective quadrupole arrangement

    bulk.cell[1] += n2u * unit_cell_u

    C1_quadrupole, C2_quadrupole, C3_quadrupole = bulk.get_cell()

    # calculation of dislocation cores positions
    # distance between centers of triangles along x
    # move to odd/even number -> get to upward/downward triangle
    x_core_dist = alat * np.sqrt(6.)/6.0

    # distance between centers of triangles along y
    y_core_dist = alat * np.sqrt(2.)/6.0

    # separation of the cores in a 1ux1v cell
    nx_left = 2
    nx_right = 5

    if n2v % 2 == 0:  # check if the number of cells in y direction is even
        # Even: then introduce cores between two equal halves of the cell
        ny_left = -2
        ny_right = -1

    else:  # Odd: introduce cores between two equal halves of the cell
        ny_left = 4
        ny_right = 5

    nx_left += 2.0 * left_shift
    nx_right += 2.0 * right_shift

    dislo_coord_left = np.array([nx_left * x_core_dist,
                                 ny_left * y_core_dist,
                                0.0])

    dislo_coord_right = np.array([nx_right * x_core_dist,
                                  ny_right * y_core_dist,
                                  0.0])

    # calculation of the shifts of the initial cores coordinates for the final
    # quadrupole arrangements

    # different x centering preferences for odd and even values
    if n2v % 2 == 0:  # check if the number of cells in y direction is even
        # Even:
        dislo_coord_left += (n2u - 1) * unit_cell_u
        dislo_coord_right += (n2u - 1 + np.trunc(n1u/2.0)) * unit_cell_u

    else:  # Odd:
        dislo_coord_left += n2u * unit_cell_u
        dislo_coord_right += (n2u + np.trunc(n1u/2.0)) * unit_cell_u

    dislo_coord_left += np.trunc(n2v/2.0) * unit_cell_v
    dislo_coord_right += np.trunc(n2v/2.0) * unit_cell_v

    u_quadrupole = dipole_displacement_angle(bulk,
                                             dislo_coord_left,
                                             dislo_coord_right)

    # get the image contribution from the dipoles around
    # (default value of N images to scan n_img=10)
    u_img = get_u_img(bulk,
                      dislo_coord_left,
                      dislo_coord_right)

    u_sum = u_quadrupole + u_img

    # calculate the field of neghbouring cell to estimate
    # linear u_err along C1 and C2 (see. Cai paper)

    # u_err along C2

    n1_shift = 0
    n2_shift = 1

    shift = n1_shift*C1_quadrupole + n2_shift*C2_quadrupole

    u_quadrupole_shifted = dipole_displacement_angle(bulk,
                                                     dislo_coord_left + shift,
                                                     dislo_coord_right + shift,
                                                     shift=shift)

    u_img_shifted = get_u_img(bulk,
                              dislo_coord_left,
                              dislo_coord_right,
                              n1_shift=n1_shift, n2_shift=n2_shift)

    u_sum_shifted = u_quadrupole_shifted + u_img_shifted

    delta_u = u_sum - u_sum_shifted

    delta_u_C2 = delta_u.T[2].mean()
    print("delta u c2: %.2f " % delta_u_C2)

    # u_err along C1

    n1_shift = 1
    n2_shift = 0

    shift = n1_shift*C1_quadrupole + n2_shift*C2_quadrupole

    u_quadrupole_shifted = dipole_displacement_angle(bulk,
                                                     dislo_coord_left + shift,
                                                     dislo_coord_right + shift,
                                                     shift=shift)

    u_img_shifted = get_u_img(bulk,
                              dislo_coord_left,
                              dislo_coord_right,
                              n1_shift=n1_shift, n2_shift=n2_shift)

    u_sum_shifted = u_quadrupole_shifted + u_img_shifted

    delta_u = u_sum - u_sum_shifted

    delta_u_C1 = delta_u.T[2].mean()
    print("delta u c1: %.3f" % delta_u_C1)

    x_scaled, y_scaled, __ = bulk.get_scaled_positions(wrap=False).T

    u_err_C2 = (y_scaled - 0.5)*delta_u_C2

    u_err_C1 = delta_u_C1*(x_scaled - 0.5)

    u_err = u_err_C1 + u_err_C2

    # Calculate the u_tilt to accomodate the stress (see. Cai paper)

    burgers = bulk.cell[2][2]

    u_tilt = 0.5 * burgers * (y_scaled - 0.5)

    final_u = u_sum
    final_u.T[2] += u_err - u_tilt

    disloc_quadrupole = bulk.copy()
    disloc_quadrupole.positions += final_u
    # tilt the cell according to the u_tilt
    disloc_quadrupole.cell[1][2] -= burgers/2.0
    bulk.cell[1][2] -= burgers/2.0

    return disloc_quadrupole, bulk, dislo_coord_left, dislo_coord_right


def make_screw_quadrupole_kink(alat, kind="double", n1u=5, kink_length=20, symbol="W"):
    """Generates kink configuration using make_screw_quadrupole() function
       works for BCC structure.
       The method is based on paper https://doi.org/10.1016/j.jnucmat.2008.12.053

    Parameters
    ----------
    alat : float
        Lattice parameter of the system in Angstrom.
    kind : string
        kind of the kink: right, left or double
    n1u : int
        Number of lattice vectors for the quadrupole cell (make_screw_quadrupole() function)
    kink_length : int
        Length of the cell per kink along b in unit of b, must be even.
    symbol : string
        Symbol of the element to pass to ase.lattuce.cubic.SimpleCubicFactory
        default is "W" for tungsten

    Returns
    -------

    kink : ase.atoms
        kink configuration
    reference_straight_disloc : ase.atoms
        reference straight dislocation configuration
    large_bulk : ase.atoms
        large bulk cell corresponding to the kink configuration

    """

    b = np.sqrt(3.0) * alat / 2.0
    cent_x = np.sqrt(6.0) * alat / 3.0

    ini_disloc_quadrupole, W_bulk, _, _ = make_screw_quadrupole(alat, n1u=n1u,
                                                                   left_shift=0.0,
                                                                   right_shift=0.0,
                                                                   symbol=symbol)

    fin_disloc_quadrupole, W_bulk, _, _ = make_screw_quadrupole(alat, n1u=n1u,
                                                                   left_shift=1.0,
                                                                   right_shift=1.0,
                                                                   symbol=symbol)

    reference_straight_disloc = ini_disloc_quadrupole * [1, 1, kink_length]
    large_bulk = W_bulk * [1, 1, kink_length]
    __, __, z = large_bulk.positions.T

    if kind == "left":

        # we have to adjust the cell to make the kink vector periodic
        # here we remove one atomic row . it is nicely explained in the paper
        left_kink_mask = z < large_bulk.get_cell()[2][2] - 1.0 * b / 3.0 - 0.01
        large_bulk.cell[2][0] -= cent_x
        large_bulk.cell[2][2] -= 1.0 * b / 3.0
        large_bulk = large_bulk[left_kink_mask]

        kink = fin_disloc_quadrupole * [1, 1, kink_length // 2]
        upper_kink = ini_disloc_quadrupole * [1, 1, kink_length // 2]
        upper_kink.positions += np.array((0.0, 0.0, kink.cell[2][2]))
        kink.cell[2][2] += upper_kink.cell[2][2]
        kink.extend(upper_kink)

        # left kink is created the kink vector is in negative x direction assuming (x, y, z) is right group of vectors
        kink = kink[left_kink_mask]
        kink.cell[2][0] -= cent_x
        kink.cell[2][2] -= 1.0 * b / 3.0

    elif kind == "right":

        # we have to adjust the cell to make the kink vector periodic
        # here we remove two atomic rows . it is nicely explained in the paper
        right_kink_mask = z < large_bulk.cell[2][2] - 2.0 * b / 3.0 - 0.01
        large_bulk.cell[2][0] += cent_x
        large_bulk.cell[2][2] -= 2.0 * b / 3.0
        large_bulk = large_bulk[right_kink_mask]

        kink = ini_disloc_quadrupole * [1, 1, kink_length // 2]
        upper_kink = fin_disloc_quadrupole * [1, 1, kink_length // 2]
        upper_kink.positions += np.array((0.0, 0.0, kink.cell[2][2]))
        kink.cell[2][2] += upper_kink.cell[2][2]
        kink.extend(upper_kink)

        kink = kink[right_kink_mask]
        # right kink is created when the kink vector is in positive x direction
        # assuming (x, y, z) is right group of vectors
        kink.cell[2][0] += cent_x
        kink.cell[2][2] -= 2.0 * b / 3.0

    elif kind == "double":

        # for the double kink it is kink length per kink
        kink = ini_disloc_quadrupole * [1, 1, kink_length // 2]
        middle_kink = fin_disloc_quadrupole * [1, 1, kink_length]
        middle_kink.positions += np.array((0.0, 0.0, kink.get_cell()[2][2]))

        kink.extend(middle_kink)
        kink.cell[2][2] += middle_kink.cell[2][2]

        upper_kink = ini_disloc_quadrupole * [1, 1, kink_length // 2]
        upper_kink.positions += np.array((0.0, 0.0, kink.get_cell()[2][2]))
        kink.extend(upper_kink)

        kink.cell[2][2] += upper_kink.cell[2][2]

        large_bulk = W_bulk * [1, 1, 2 * kink_length]  # double kink is double length

    else:
        raise ValueError('Kind must be "right", "left" or "double"')

    return kink, reference_straight_disloc, large_bulk


def make_edge_cyl_001_100(a0, C11, C12, C44,
                          cylinder_r,
                          cutoff=5.5,
                          tol=1e-6,
                          symbol="W"):
    """Function to produce consistent edge dislocation configuration.

    Parameters
    ----------
    alat : float
        Lattice constant of the material.
    C11 : float
        C11 elastic constant of the material.
    C12 : float
        C12 elastic constant of the material.
    C44 : float
        C44 elastic constant of the material.
    cylinder_r : float
        Radius of cylinder of unconstrained atoms around the
        dislocation  in angstrom.
    cutoff : float
        Potential cutoff for determenition of size of
        fixed atoms region (2*cutoff)
    tol : float
        Tolerance for generation of self consistent solution.
    symbol : string
        Symbol of the element to pass to ase.lattuce.cubic.SimpleCubicFactory
        default is "W" for tungsten

    Returns
    -------
    bulk : ase.Atoms object
        Bulk configuration.
    disloc : ase.Atoms object
        Dislocation configuration.
    disp : np.array
        Corresponding displacement.
    """
    # https://github.com/usnistgov/atomman
    from atomman import ElasticConstants
    from atomman.defect import Stroh
    # Create a Stroh ojbect with junk data
    stroh = Stroh(ElasticConstants(C11=141, C12=110, C44=98),
                  np.array([0, 0, 1]))

    axes = np.array([[1, 0, 0],
                     [0, 1, 0],
                     [0, 0, 1]])

    c = ElasticConstants(C11=C11, C12=C12, C44=C44)
    burgers = a0 * np.array([1., 0., 0.])

    # Solving a new problem with Stroh.solve
    # Does not work with the new version of atomman
    stroh.solve(c, burgers, axes=axes)

    unit_cell = BodyCenteredCubic(directions=axes.tolist(),
                                  size=(1, 1, 1), symbol=symbol,
                                  pbc=(False, False, True),
                                  latticeconstant=a0)

    bulk = unit_cell.copy()

    # shift to make the zeros of the cell betweem the atomic planes
    # and under the midplane on Y axes
    X_midplane_shift = -0.25*a0
    Y_midplane_shift = -0.25*a0

    bulk_shift = [X_midplane_shift,
                  Y_midplane_shift,
                  0.0]

    bulk.positions += bulk_shift

    tot_r = cylinder_r + 2*cutoff + 0.01

    Lx = int(round(tot_r/a0))
    Ly = int(round(tot_r/a0))

    # factor 2 to make sure odd number of images is translated
    # it is important for the correct centering of the dislocation core
    bulk = bulk * (2*Lx, 2*Ly, 1)

    center_shift = [Lx * a0, Ly * a0, 0.0]

    bulk.positions -= center_shift
    # bulk.write("before.xyz")

    disp1 = stroh.displacement(bulk.positions)
    disloc = bulk.copy()
    res = np.inf
    i = 0
    while res > tol:
        disloc.positions = bulk.positions + disp1
        disp2 = stroh.displacement(disloc.positions)
        res = np.abs(disp1 - disp2).max()
        disp1 = disp2
        print('disloc SCF', i, '|d1-d2|_inf =', res)
        i += 1
        if i > 10:
            raise RuntimeError('Self-consistency did \
                                not converge in 10 cycles')
    disp = disp2

    x, y, z = disloc.positions.T
    radius_x_y_zero = np.sqrt(x**2 + y**2)
    mask = radius_x_y_zero < tot_r
    disloc = disloc[mask]
    bulk = bulk[mask]
    # disloc.write("after_disp.xyz")

    x, y, z = disloc.positions.T
    radius_x_y_zero = np.sqrt(x**2 + y**2)
    mask_zero = radius_x_y_zero > cylinder_r
    fix_atoms = FixAtoms(mask=mask_zero)
    disloc.set_constraint(fix_atoms)

    return bulk, disloc, disp


def read_dislo_QMMM(filename=None, image=None):

    """Reads extended xyz file with QMMM configuration
       Uses "region" for mapping of QM, MM and fixed atoms
       Sets ase.constraints.FixAtoms constraint on fixed atoms

    Parameters
    ----------
    filename : path to xyz file
    image : image with "region" array to set up constraint and extract qm_mask

    Returns
    -------
    dislo_QMMM : Output ase.Atoms object
        Includes "region" array and FixAtoms constraint
    qm_mask : array mask for QM atoms mapping
    """

    if filename is not None:
        dislo_QMMM = read(filename)

    elif image is not None:
        dislo_QMMM = image

    else:
        raise RuntimeError("Please provide either path or image")

    region = dislo_QMMM.get_array("region")
    Nat = len(dislo_QMMM)

    print("Total number of atoms in read configuration: {0:7}".format(Nat))

    for region_type in np.unique(region):
        print("{0:52d} {1}".format(np.count_nonzero(region == region_type),
                                   region_type))

    if len(dislo_QMMM.constraints) == 0:

        print("Adding fixed atoms constraint")

        fix_mask = region == "fixed"
        fix_atoms = FixAtoms(mask=fix_mask)
        dislo_QMMM.set_constraint(fix_atoms)

    else:
        print("Constraints list is not zero")

    qm_mask = region == "QM"

    qm_atoms = dislo_QMMM[qm_mask]
    qm_atoms_types = np.array(qm_atoms.get_chemical_symbols())

    print("QM region atoms: {0:3d}".format(np.count_nonzero(qm_mask)))

    for qm_atom_type in np.unique(qm_atoms_types):

        print("{0:20d} {1}".format(np.count_nonzero(qm_atoms_types == qm_atom_type),
                                   qm_atom_type))

    return dislo_QMMM, qm_mask



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2019) Johannes Hoermann, University of Freiburg
#                  Lukas Elflein, University of Freiburg
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# ======================================================================
"""
Generates atomic structure following a given distribution.

Copyright 2019 IMTEK Simulation
University of Freiburg

Authors:

  Johannes Hoermann <johannes.hoermann@imtek-uni-freiburg.de>
  Lukas Elflein <elfleinl@cs.uni-freiburg.de>
"""
import logging, os, sys
import os.path
from six.moves import builtins
from collections.abc import Iterable

import numpy as np
# import matplotlib.pyplot as plt

import scipy.constants as sc
from scipy import integrate, optimize

logger = logging.getLogger(__name__)

def exponential(x, rate=0.1):
    """Exponential distribution."""
    return rate * np.exp(-1 * rate * x)


def uniform(x, *args, **kwargs):
    """Uniform distribution."""
    return np.ones(np.array(x).shape) / 2

def pdf_to_cdf(pdf):
    """Transform partial distribution to cumulative distribution function

    >>> pdf_to_cdf(pdf=np.array([0.5, 0.3, 0.2]))
    array([ 0.5,  0.8,  1. ])
    """
    cdf = np.cumsum(pdf)
    cdf /= cdf[-1]

    return cdf


def get_centers(bins):
    """Return the center of the provided bins.

    Example:
    >>> get_centers(bins=np.array([0.0, 1.0, 2.0]))
    array([ 0.5,  1.5])
    """
    bins = bins.astype(float)
    return (bins[:-1] + bins[1:]) / 2


def get_nearest_pos(array, value):
    """Find the value of an array clostest to the second argument.

    Example:
    >>> get_nearest_pos(array=[0, 0.25, 0.5, 0.75, 1.0], value=0.55)
    2
    """
    array = np.asarray(array)
    pos = np.abs(array - value).argmin()
    return pos


def get_histogram(struc, box, n_bins=100):
    """Slice the list of atomic positions, aggregate positions into histogram."""
    # Extract x/y/z positions only
    # x, y, z = struc[:, 0], struc[:, 1], struc[:, 2]

    histograms = []
    for dimension in range(struc.shape[1]):
        bins = np.linspace(0, box[dimension], n_bins)
        hist, bins = np.histogram(struc[:, dimension], bins=bins, density=True)
        # Normalize the histogram for all values to sum to 1
        hist /= sum(hist)

        histograms += [(hist, bins)]
    return histograms


# def plot_dist(histogram, name, reference_distribution=None):
#     """Plot histogram with an optional reference distribution."""
#     hist, bins = histogram
#     width = 1 * (bins[1] - bins[0])
#     centers = get_centers(bins)
#
#     fi, ax = plt.subplots()
#     ax.bar( centers, hist, align='center', width=width, label='Empirical distribution',
#             edgecolor="none")
#
#     if reference_distribution is not None:
#         ref = reference_distribution(centers)
#         ref /= sum(ref)
#         ax.plot(centers, ref, color='red', label='Target distribution')
#
#     plt.title(name)
#     plt.legend()
#     plt.xlabel('Distance ' + name)
#     plt.savefig(name + '.png')


def quartile_function(distribution, p, support=None):
    """Inverts a distribution x->p, and returns the x-value belonging to the provided p.

    Assumption: The distribution to be inverted must have a strictly increasing CDF!
    Also see 'https://en.wikipedia.org/wiki/Quantile_function'.

    Parameters
    ----------
    distribution: a function x -> p; x should be approximatable by a compact support
    p: an output of the distribution function, probablities in (0,1) are preferrable
    """
    if support is None:
        # Define the x-values to evaluate the function on
        support = np.arange(0,1,0.01)

    # Calculate the histogram of the distribution
    hist = distribution(support)

    # Sum the distribution to get the cumulatative distribution
    cdf =  pdf_to_cdf(hist)

    # If the p is not in the image of the support, get the nearest hit instead
    nearest_pos = get_nearest_pos(cdf, p)

    # Get the x-value belonging to the probablity value provided in the input
    x = support[nearest_pos]
    return x

def inversion_sampler(distribution, support):
    """Wrapper for quartile_function."""
    # z is distributed according to the given distribution
    # To approximate this, we insert an atom with probablity dis(z) at place z.
    # This we do by inverting the distribution, and sampling uniformely from distri^-1:
    p = np.random.uniform()
    sample = quartile_function(distribution, p, support=support)

    return sample


def rejection_sampler(distribution, support=(0.0,1.0), max_tries=10000, scale_M=1.1):
    """Sample distribution by drawing from support and keeping according to distribution.

    Draw a random sample from our support, and keep it if another random number is
    smaller than our target distribution at the support location.

        Algorithm: https://en.wikipedia.org/wiki/Rejection_sampling

    Parameters
    ----------
    distribution: callable(x)
        target distribut10on
    support: list or 2-tuple
        either discrete list of locations in space where our distribution is
        defined, or 2-tuple defining conitnuous support interval
    max_tries: how often the sampler should attempt to draw before giving up.
       If the distribution is very sparse, increase this parameter to still get results.
    scale_M: float, optional
        scales bound M for likelihood ratio

    Returns
    -------
    sample: a location which is conistent (in expectation) with being drawn from the distribution.
    """

    # rejection sampling (https://en.wikipedia.org/wiki/Rejection_sampling):
    # Generates sampling values from a target distribution X with arbitrary
    # probability density function f(x) by using a proposal distribution Y
    # with probability density g(x).
    # Concept: Generates a sample value from X by instead sampling from Y and
    # accepting this sample with probability f(x) / ( M g(x) ), repeating the
    # draws from Y until a value is accepted. M here is a constant, finite bound
    # on the likelihood ratio f(x)/g(x), satisfying 1 < M < infty over the
    # support of X; in other words, M must satisfy f(x) <= Mg(x) for all values
    # of x. The support of Y must include the support of X.

    # Here, f(x) = distribution(x), g(x) is uniform density on [0,1)
    # X are f-distributed positions from support, Y are uniformly distributed
    # values from [0,1)]
    logger.debug("Rejection sampler on distribution f(x) ({}) with".format(
        distribution))

    # coninuous support case
    if isinstance(support,tuple) and len(support) == 2:
        a = support[0]
        b = support[1]
        logger.debug("continuous support X (interval [{},{}]".format(a,b))
        # uniform probability density g(x) on support is
        g = 1 / ( b - a )
        # find maximum value fmax on distribution at x0
        xatol = (b - a)*1e-6 # optimization absolute tolerance
        x0 = optimize.minimize_scalar( lambda x: -distribution(x),
            bounds=(a,b), method='bounded', options={'xatol':xatol}).x
        fmax = distribution(x0)
        M = scale_M*fmax / g
        logger.debug("Uniform probability density g(x) = {:g} and".format(g))
        logger.debug("maximum probability density f(x0) = {:g} at x0 = {:g}".format(fmax, x0))
        logger.debug("require M >= scale_M*g(x)/max(f(x)), i.e. M = {:g}.".format(M))

        for i in range(max_tries):
            # draw a sample from a uniformly distributed support
            sample = np.random.random() * (b-a) + a

            # Generate random float in the half-open interval [0.0, 1.0) and .
            # keep sample with probablity of distribution
            if np.random.random() < distribution(sample) / (M*g):
                return sample

    else: # discrete support case
        logger.debug("discrete support X ({:d} points in interval [{},{}]".format(
            len(support), np.min(support), np.max(support)))
        # uniform probability density g(x) on support is
        g = 1.0 / len(support) # for discrete support
        # maximum probability on distributiom f(x) is
        fmax = np.max(distribution(support))
        # thus M must be at least
        M = scale_M * fmax / g
        logger.debug("Uniform probability g(x) = {:g} and".format(g))
        logger.debug("maximum probability max(f(x)) = {:g} require".format(fmax))
        logger.debug("M >= scale_M*g(x)/max(f(x)), i.e. M = {:g}.".format(M))

        for i in range(max_tries):
            # draw a sample from support
            sample = np.random.choice(support)

            # Generate random float in the half-open interval [0.0, 1.0) and .
            # keep sample with probablity of distribution
            if np.random.random() < distribution(sample) / (M*g):
                return sample

    raise RuntimeError('Maximum of attempts max_tries {} exceeded!'.format(max_tries))


def generate_structure(
    distribution, box=np.array([50, 50, 100]),
    count=100, n_gridpoints=np.nan):
    """Generate 'atoms' from continuous distributuion(s).

    Coordinates are distributed according to given distributions.

    Per default, X and Y coordinates are drawn uniformely.

    Parameters
    ----------
    distribution: func(x) or list of func(x)
      With one function, uniform sampling appplies along x and y axes,
      while applying 'distribution' along z axis. With a list of functions,
      apllies the respective distribution function along x, y and z direction.
    box: np.ndarray(3), optional (default: np.array([50, 50, 100]) )
      dimensions of volume to be filled with samples
    count: int, optional (default: 100)
      number of samples to draw
    n_gridpoints: int or (int,int,int), optional (default: np.nan)
      If spcefified, samples are not placed arbitrarily, but on an evenly spaced
      grid of this many grid points along each axis. Specify np.nan for
      continuous sampling, i.e. (10,np.nan,20)

    Returns
    -------
    np.ndarray((sample_size,3)): sample coordinates
    """
    global logger

    if callable(distribution):
        logger.info("Using uniform distribution along x and y direction.")
        logger.info("Using distribution {} along z direction.".format(
            distribution))
        distribution = [ uniform, uniform, distribution]

    for d in distribution:
        assert callable(d), "distribution {} must be callable".format(d)

    assert np.array(box).shape == (3,), "wrong specification of 3d box dimensions"

    #if isinstance(n_gridpoints,int) or n_gridpoints == np.nan:
    if not isinstance(n_gridpoints, Iterable):
        n_gridpoints = 3*[n_gridpoints] # to list

    n_gridpoints = np.array(n_gridpoints,dtype=float)
    logger.info("Using {} grid as sampling support.".format(
        n_gridpoints))

    assert n_gridpoints.shape == (3,), "n_gridpoints must be int or list of int"

    # We define which positions in space the atoms can be placed
    support = []
    normalized_distribution = []
    for k, d in enumerate(distribution):
        # Using the box parameter, we construct a grid inside the box
        # This results in a 100x100x100 grid:
        if np.isnan( n_gridpoints[k] ): # continuous support
            support.append((0,box[k])) # interval
            # Normalization constant:
            Z, _ = integrate.quad(d, support[-1][0], support[-1][1])
        else : # discrete supoport
            support.append(np.linspace(0, box[k], n_gridpoints[k]))
            Z = np.sum(d(support[-1])) # Normalization constant

        logger.info("Normalizing 'distribution' {} by {}.".format(d,Z))
        normalized_distribution.append(
            lambda x,k=k,Z=Z: distribution[k](x) / Z )

    # For every atom, draw random x, y and z coordinates
    positions = np.array( [ [
        rejection_sampler(d,s) for d,s in zip(normalized_distribution,support) ]
            for i in range(int(count)) ])

    logger.info("Drew {} samples from distributions.".format(positions.shape))
    return positions



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

import itertools

import numpy as np

from ase.data import atomic_numbers

import _matscipy
from _matscipy import first_neighbours

###

def mic(dr, cell, pbc=None):
    """
    Apply minimum image convention to an array of distance vectors.

    Parameters
    ----------
    dr : array_like
        Array of distance vectors.
    cell : array_like
        Simulation cell.
    pbc : array_like, optional
        Periodic boundary conditions in x-, y- and z-direction. Default is to
        assume periodic boundaries in all directions.

    Returns
    -------
    dr : array
        Array of distance vectors, wrapped according to the minimum image
        convention.
    """
    # Check where distance larger than 1/2 cell. Particles have crossed
    # periodic boundaries then and need to be unwrapped.
    rec = np.linalg.inv(cell)
    if pbc is not None:
        rec *= np.array(pbc, dtype=int).reshape(3,1)
    dri = np.round(np.dot(dr, rec))

    # Unwrap
    return dr - np.dot(dri, cell)


def neighbour_list(quantities, atoms=None, cutoff=None, positions=None,
                   cell=None, pbc=None, numbers=None, cell_origin=None):
    """
    Compute a neighbor list for an atomic configuration. Atoms outside periodic
    boundaries are mapped into the box. Atoms outside nonperiodic boundaries
    are included in the neighbor list but the complexity of neighbor list search
    for those can become n^2.

    The neighbor list is sorted by first atom index 'i', but not by second 
    atom index 'j'.

    The neighbour list accepts either an ASE Atoms object or positions and cell
    vectors individually.

    Parameters
    ----------
    quantities : str
        Quantities to compute by the neighbor list algorithm. Each character
        in this string defines a quantity. They are returned in a tuple of
        the same order. Possible quantities are
            'i' : first atom index
            'j' : second atom index
            'd' : absolute distance
            'D' : distance vector
            'S' : shift vector (number of cell boundaries crossed by the bond
                  between atom i and j). With the shift vector S, the
                  distances D between atoms can be computed from:
                  D = a.positions[j]-a.positions[i]+S.dot(a.cell)
    atoms : ase.Atoms
        Atomic configuration. (Default: None)
    cutoff : float or dict
        Cutoff for neighbor search. It can be
            - A single float: This is a global cutoff for all elements.
            - A dictionary: This specifies cutoff values for element
              pairs. Specification accepts element numbers of symbols.
              Example: {(1, 6): 1.1, (1, 1): 1.0, ('C', 'C'): 1.85}
            - A list/array with a per atom value: This specifies the radius of
              an atomic sphere for each atoms. If spheres overlap, atoms are
              within each others neighborhood.
    positions : array_like
        Atomic positions. (Default: None)
    cell : array_like
        Cell vectors as a 3x3 matrix. (Default: Shrink wrapped cell)
    pbc : array_like
        3-vector containing periodic boundary conditions in all three
        directions. (Default: Nonperiodic box)
    numbers : array_like
        Array containing the atomic numbers.

    Returns
    -------
    i, j, ... : array
        Tuple with arrays for each quantity specified above. Indices in `i`
        are returned in ascending order 0..len(a), but the order of (i,j)
        pairs is not guaranteed.

    Examples
    --------
    Examples assume Atoms object *a* and numpy imported as *np*.
    1. Coordination counting:
        i = neighbor_list('i', a, 1.85)
        coord = np.bincount(i)

    2. Coordination counting with different cutoffs for each pair of species
        i = neighbor_list('i', a,
                           {('H', 'H'): 1.1, ('C', 'H'): 1.3, ('C', 'C'): 1.85})
        coord = np.bincount(i)

    3. Pair distribution function:
        d = neighbor_list('d', a, 10.00)
        h, bin_edges = np.histogram(d, bins=100)
        pdf = h/(4*np.pi/3*(bin_edges[1:]**3 - bin_edges[:-1]**3)) * a.get_volume()/len(a)

    4. Pair potential:
        i, j, d, D = neighbor_list('ijdD', a, 5.0)
        energy = (-C/d**6).sum()
        pair_forces = (6*C/d**5  * (D/d).T).T
        forces_x = np.bincount(j, weights=pair_forces[:, 0], minlength=len(a)) - \
                   np.bincount(i, weights=pair_forces[:, 0], minlength=len(a))
        forces_y = np.bincount(j, weights=pair_forces[:, 1], minlength=len(a)) - \
                   np.bincount(i, weights=pair_forces[:, 1], minlength=len(a))
        forces_z = np.bincount(j, weights=pair_forces[:, 2], minlength=len(a)) - \
                   np.bincount(i, weights=pair_forces[:, 2], minlength=len(a))

    5. Dynamical matrix for a pair potential stored in a block sparse format:
        from scipy.sparse import bsr_matrix
        i, j, dr, abs_dr = neighbor_list('ijDd', atoms)
        energy = (dr.T / abs_dr).T
        dynmat = -(dde * (energy.reshape(-1, 3, 1) * energy.reshape(-1, 1, 3)).T).T \
                 -(de / abs_dr * (np.eye(3, dtype=energy.dtype) - \
                   (energy.reshape(-1, 3, 1) * energy.reshape(-1, 1, 3))).T).T
        dynmat_bsr = bsr_matrix((dynmat, j, first_i), shape=(3*len(a), 3*len(a)))

        dynmat_diag = np.empty((len(a), 3, 3))
        for x in range(3):
            for y in range(3):
                dynmat_diag[:, x, y] = -np.bincount(i, weights=dynmat[:, x, y])

        dynmat_bsr += bsr_matrix((dynmat_diag, np.arange(len(a)),
                                  np.arange(len(a) + 1)),
                                 shape=(3 * len(a), 3 * len(a)))


  i_n, j_n, dr_nc, abs_dr_n = neighbour_list('ijDd', atoms, dict)
e_nc = (dr_nc.T/abs_dr_n).T
    D_ncc = -(dde_n * (e_nc.reshape(-1,3,1) * e_nc.reshape(-1,1,3)).T).T
    D_ncc += -(de_n/abs_dr_n * (np.eye(3, dtype=e_nc.dtype) - (e_nc.reshape(-1,3,1) * e_nc.reshape(-1,1,3))).T).T

    D = bsr_matrix((D_ncc, j_n, first_i), shape=(3*nat,3*nat))

    Ddiag_icc = np.empty((nat,3,3))
    for x in range(3):
        for y in range(3):
            Ddiag_icc[:,x,y] = -np.bincount(i_n, weights = D_ncc[:,x,y])

    D += bsr_matrix((Ddiag_icc,np.arange(nat),np.arange(nat+1)), shape=(3*nat,3*nat))

    return D
    """

    if cutoff is None:
        raise ValueError('Please provide a value for the cutoff radius.')

    if atoms is None:
        if positions is None:
            raise ValueError('You provided neither an ASE Atoms object nor '
                             'a positions array.') 
        if cell is None:
            # Shrink wrapped cell
            rmin = np.min(positions, axis=0)
            rmax = np.max(positions, axis=0)
            cell_origin = rmin
            cell = np.diag(rmax - rmin)
        if cell_origin is None:
            cell_origin = np.zeros(3)
        if pbc is None:
            pbc = np.zeros(3, dtype=bool)
        if numbers is None:
            numbers = np.ones(len(positions), dtype=np.int32)
    else:
        if positions is not None:
            raise ValueError('You cannot provide an ASE Atoms object and '
                             'individual position atomic positions at the same '
                             'time.')
        positions = atoms.positions
        if cell_origin is not None:
            raise ValueError('You cannot provide an ASE Atoms object and '
                             'a cell origin at the same time.')
        cell_origin = np.zeros(3)
        if cell is not None:
            raise ValueError('You cannot provide an ASE Atoms object and '
                             'cell vectors at the same time.')
        cell = atoms.cell
        if pbc is not None:
            raise ValueError('You cannot provide an ASE Atoms object and '
                             'separate periodicity information at the same '
                             'time.')
        pbc = atoms.pbc
        if numbers is not None:
            raise ValueError('You cannot provide an ASE Atoms object and '
                             'separate atomic numbers at the same time.')
        numbers = atoms.numbers.astype(np.int32)

    if isinstance(cutoff, dict):
        maxel = np.max(numbers)
        _cutoff = np.zeros([maxel+1, maxel+1], dtype=float)
        for (el1, el2), c in cutoff.items():
            try:
                el1 = atomic_numbers[el1]
            except:
                pass
            try:
                el2 = atomic_numbers[el2]
            except:
                pass
            if el1 < maxel+1 and el2 < maxel+1:
                _cutoff[el1, el2] = c
                _cutoff[el2, el1] = c
    else:
        _cutoff = cutoff

    return _matscipy.neighbour_list(quantities, cell_origin, cell,
                                    np.linalg.inv(cell.T), pbc, positions,
                                    _cutoff, numbers)


def find_indices_of_reversed_pairs(i_n, j_n, abs_dr_n):
    """Find neighbor list indices where reversed pairs are stored

    Given list of identifiers of neighbor atoms `i_n` and `j_n`,
    determines the list of indices `reverse` into the neighbor list
    where each pair is reversed, i.e. `i_n[reverse[n]]=j_n[n]` and
    `j_n[reverse[n]]=i_n[n]` for each index `n` in the neighbor list
    
    In the case of small periodic systems, one needs to be careful, because
    the same pair may appear more than one time, with different pair
    distances. Therefore, the pair distance must be taken into account.

    We assume that there is in fact one reversed pair for every pair.
    However, we do not check this assumption in order to avoid overhead.

    Parameters
    ----------
    i_n : array_like
       array of atom identifiers
    j_n : array_like
       array of atom identifiers
    abs_dr_n : array_like
        pair distances

    Returns
    -------
    reverse : numpy.ndarray
        array of indices into i_n and j_n
    """
    sorted_1 = np.lexsort(keys=(abs_dr_n, i_n, j_n))
    sorted_2 = np.lexsort(keys=(abs_dr_n, j_n, i_n))
    #np.testing.assert_equal(j_n[sorted_1], i_n[sorted_2])
    #np.testing.assert_equal(i_n[sorted_1], j_n[sorted_2])
    #np.testing.assert_equal(abs_dr_n[sorted_1], abs_dr_n[sorted_2])
    #print(np.c_[i_n[sorted_2], j_n[sorted_2], abs_dr_n[sorted_2], 
    #            i_n[sorted_1], j_n[sorted_1], abs_dr_n[sorted_1]])
    tmp2 = np.arange(i_n.size)[sorted_2]
    tmp1 = np.arange(i_n.size)[sorted_1]
    # np.arange(i_n.size) are indices into the neighbor list, so
    #  * the nth element in tmp1 is the index where i,j was before reordering with sorted_1
    #  * the nth element in tmp2 is the index where j,i was before reordering with sorted_2
    reverse = np.empty(i_n.size, dtype=i_n.dtype)
    reverse[tmp1] = tmp2
    return reverse



import os
import shutil
import subprocess
import socket
import socketserver
from io import StringIO
import time
import threading
from queue import Queue

import numpy as np

from matscipy.elasticity import full_3x3_to_Voigt_6_stress
from matscipy.logger import quiet, screen

from ase.atoms import Atoms
from ase.io.extxyz import read_xyz, write_xyz
from ase.io.vasp import write_vasp
from ase.io.castep import write_castep_cell, write_param

from ase.calculators.calculator import Calculator
from ase.calculators.vasp import Vasp
from ase.calculators.castep import Castep

MSG_LEN_SIZE = 8
MSG_END_MARKER = b'done.\n'
MSG_END_MARKER_SIZE = len(MSG_END_MARKER)
MSG_INT_SIZE = 6
MSG_FLOAT_SIZE = 25
MSG_FLOAT_FORMAT = '%25.16f'
MSG_INT_FORMAT = '%6d'

ATOMS_REQUESTS = {ord('A'): 'REFTRAJ', ord('X'): 'XYZ'}
RESULTS_REQUESTS = {ord('R'): 'REFTRAJ', ord('Y'): 'XYZ'}
ZERO_ATOMS_DATA = {'REFTRAJ': b'     242     0\n     0\n       0.0000000000000000       0.0000000000000000       0.0000000000000000\n       0.0000000000000000       0.0000000000000000       0.0000000000000000\n       0.0000000000000000       0.0000000000000000       0.0000000000000000\n',
                   'XYZ': b'     2500\nlabel=0 cutoff_factor=1.20000000 nneightol=1.20000000 Lattice="0.00000000       0.00000000       0.00000000       0.00000000       0.00000000       0.00000000       0.00000000       0.00000000       0.00000000" Properties=species:S:1:pos:R:3:Z:I:1\n'}
CLIENT_TIMEOUT = 60
MAX_POS_DIFF = 1.0   # angstrom
MAX_CELL_DIFF = 1e-3 # angstrom

MAX_POS_DIFF_CASTEP = 1.0  # angstrom
MAX_CELL_DIFF_CASTEP = 1.0 # angstrom

def pack_atoms_to_reftraj_str(at, label):
    data = ''
    data += MSG_INT_FORMAT % label + '\n'
    data += MSG_INT_FORMAT % len(at) + '\n'
    for i in range(3):
        data += (3*MSG_FLOAT_FORMAT) % tuple(at.cell[:, i]) + '\n'
    s = at.get_scaled_positions()
    for i in range(len(at)):
        data += (3*MSG_FLOAT_FORMAT) % tuple(s[i, :]) + '\n'
    # preceed message by its length
    data_length = ('%8d' % len(data)).encode('ascii')
    data = data_length + data.encode('ascii')
    return data

def pack_atoms_to_xyz_str(at, label):
    at.info['label'] = label
    buffer = StringIO.StringIO()
    write_xyz(buffer, at)
    data = str(buffer)
    buffer.close()
    # preceed message by its length
    data_length = ('%8d' % len(data)).encode('ascii')
    data = data_length + data.encode('ascii')
    return data

def unpack_reftraj_str_to_atoms(data):
    lines = data.split(b'\n')
    label = int(lines[0])
    n_atoms = int(lines[1])
    at = Atoms(symbols=[' ']*n_atoms, cell=np.eye(3))
    at.info['label'] = label
    for i in range(3):
        at.cell[:, i] = [float(x) for x in lines[i].split()]
    for i, line in enumerate(lines[4:]):
        t = [float(x) for x in line.split()]
        at.positions[i, :] = np.dot(t, at.cell)
    return at

def pack_results_to_reftraj_output_str(at):
    data = ''
    data += MSG_INT_FORMAT % len(at) + '\n'
    data += MSG_FLOAT_FORMAT % at.energy + '\n'
    force = at.get_array('force')
    virial = at.info['virial']
    for i in at.indices:
        data += (3*MSG_FLOAT_FORMAT) % tuple(force[i, :]) + '\n'
    # NB: not in Voigt order (xx, yy, zz, yz, xz, xy)
    data += (6*MSG_FLOAT_FORMAT) % (virial[0,0], virial[1,1], virial[2,2],
                                    virial[0,1], virial[1,2], virial[0,2])

    # preceed message by its length
    data_length = ('%8s' % len(data)).encode('ascii')
    data = data_length + data
    return data

def unpack_reftraj_output_str_to_results(data):
    lines = data.strip().split(b'\n')
    label = int(lines[0])
    natoms = int(lines[1])
    energy = float(lines[2])
    force = np.zeros((natoms,3))
    for i, line in enumerate(lines[3:-1]):
       force[i, :] = [float(f) for f in line.split()]
    v6 = [float(v) for v in lines[-1].split()]
    virial = np.zeros((3,3))
    # NB: not in Voigt order (xx, yy, zz, yz, xz, xy)
    virial[0,0], virial[1,1], virial[2,2], virial[0,1], virial[1,2], virial[0,2] = v6
    virial[1,0] = virial[0,1]
    virial[2,1] = virial[1,2]
    virial[2,0] = virial[0,2]
    return (label, (natoms, energy, force, virial))

def unpack_xyz_str_to_results(data):
    buffer = StringIO.StringIO(data)
    at = read_xyz(buffer)
    buffer.close()
    label = at.info['label']
    return (label, at)

class AtomsRequestHandler(socketserver.StreamRequestHandler):
    def handle(self):
        ip, port = self.client_address
        task = None
        # receive request code and client ID
        request_str = self.rfile.read(MSG_LEN_SIZE)
        request = request_str[0]
        client_id = int(request_str[1:])

        if client_id > self.server.njobs-1:
            raise RuntimeError('Unknown client ID %d outside of range 0 < ID < %d' %
                               (client_id, self.server.njobs-1))

        self.server.logger.pr('"%s" request from %s:%d client %d' % (chr(request), ip, port, client_id))
        #print 'input queue lengths ', ''.join(['%d:%d ' % (i,q.qsize()) for (i,q) in enumerate(input_qs)])
        #print 'output queue length %d' % output_q.qsize()

        if request in ATOMS_REQUESTS:
            # client is ready for Atoms (in either REFTRAJ or XYZ format)
            data, fmt, label, at = self.server.input_qs[client_id].get()
            assert ATOMS_REQUESTS[request] == fmt
            if data == b'shutdown' or data == b'restart':
                task = data
                data = ZERO_ATOMS_DATA[fmt]
            self.wfile.write(data)

        elif request in RESULTS_REQUESTS:
            # results are available from client in REFTRAJ or XYZ format
            data_size = int(self.rfile.read(MSG_LEN_SIZE))
            data = self.rfile.read(data_size)
            fmt = RESULTS_REQUESTS[request]
            self.server.output_q.put((client_id, fmt, data))
            self.server.input_qs[client_id].task_done()

        else:
            raise RuntimeError('Unknown request code "%s"' % request)

        # say goodbye to this client
        self.wfile.write(MSG_END_MARKER)

        if (request == ord('A') or request == ord('X')) and task == b'restart':
            # if we're restarting a client, get the next thing out of the queue
            # and re-initialise. Restart won't do anything until shutdown
            # of old client has completed.
            data, fmt, label, at = self.server.input_qs[client_id].get()
            self.server.logger.pr('"%s" request from client %d triggering restart for calculation with label %d' %
                                (request, client_id, label))
            self.server.clients[client_id].start_or_restart(at, label, restart=True)


class AtomsServerSync(socketserver.TCPServer):
    allow_reuse_address = True

    def __init__(self, server_address, RequestHandlerClass, clients,
                 bind_and_activate=True, max_attempts=3, bgq=False, logger=screen):

        self.njobs = len(clients)
        # allow up to twice as many threads as sub-block jobs
        self.request_queue_size = 2*self.njobs
        self.max_attempts = max_attempts
        self.bgq = bgq # If True, we're running on IBM Blue Gene/Q platform
        self.logger = logger

        socketserver.TCPServer.__init__(self,
                                        server_address,
                                        RequestHandlerClass,
                                        bind_and_activate)

        self.clients = clients
        for client in self.clients:
            client.server = self # FIXME circular reference

        # record all input in the order in which it is put()
        self.input_q = Queue()
        # we also need an input Queue for each client: this is so that we can
        # exploit wavefunction reuse by sending consecutive clusters belonging
        # to the same atom to the same QM partition
        self.input_qs = [Queue() for i in range(self.njobs) ]
        self.output_q = Queue()



    def server_activate(self):
        socketserver.TCPServer.server_activate(self)
        self.ip, self.port = self.server_address
        if self.bgq:
            # If we're on a Blue Gene, note that IP address returned
            # by server.server_address is not the correct one for CNs
            # to talk to FEN, so we discard it, and use the InfiniBand
            # address returned by get_hostname_ip()
            import bgqtools
            hostname, self.ip = bgqtools.get_hostname_ip()
        else:
            hostname = socket.gethostname()
        self.logger.pr('AtomsServer running on %s %s:%d with njobs=%d' %
                     (hostname, self.ip, self.port, self.njobs))


    def shutdown_clients(self):
        self.logger.pr('shutting down all clients')
        wait_threads = []
        for client_id, client in enumerate(self.clients):
            if (client.process is not None and client.process.poll() is None and
                (client.wait_thread is None or not client.wait_thread.isAlive())):
                wait_threads.append(client.shutdown(block=False))
                self.handle_request() # dispatch the shutdown request via socket
        # wait for them all to finish shutting down
        for wait_thread in wait_threads:
            if wait_thread is None or not wait_thread.isAlive():
                continue
            wait_thread.join()
        self.logger.pr('all client shutdowns complete')


    def shutdown(self):
        self.shutdown_clients()
        self.server_close()


    def put(self, at, client_id, label, force_restart=False):
        self.logger.pr('Putting Atoms to client %d label %d' % (client_id, label))

        # allow client to modify atoms (e.g. sort them)
        at, fmt, first_time = self.clients[client_id].preprocess(at, label, force_restart)

        # store what we actually did- `at` may have been modified by preprocess()
        self.input_q.put((label, client_id, at))

        if fmt == 'REFTRAJ':
            data = pack_atoms_to_reftraj_str(at, label)
        elif fmt == 'XYZ':
            data = pack_atoms_to_xyz_str(at, label)
        else:
            raise ValueError('Unknown format "%s"' % fmt)
        self.input_qs[client_id].put((data, fmt, label, at))

        if first_time:
            # throw away what we just put(), as it's in the input files.
            # note that we don't call task_done() until results come in
            discard = self.input_qs[client_id].get()


    def join_all(self):
        self.logger.pr('AtomsServer waiting for input queues to empty')
        for input_q in self.input_qs:
            input_q.join()
        self.logger.pr('all AtomsServer queues drained.')


    def get_results(self):
        self.logger.pr('AtomsServer getting results')

        results = {}

        for attempt in range(self.max_attempts):
            rejects = []
            self.join_all()
            self.logger.pr('AtomsServer.get_results() attempt %d of %d jobs finished' %
                         (attempt+1, self.max_attempts))

            while self.output_q.unfinished_tasks:
                client_id, fmt, data = self.output_q.get()

                if fmt == 'REFTRAJ':
                    label, res = unpack_reftraj_output_str_to_results(data)
                elif fmt == 'XYZ':
                    label, res = unpack_xyz_str_to_results(data)
                else:
                    raise ValueError('get_results() got unknown format "%s"' % fmt)

                if label > 0: # WARNING: labels must start from 1, or first calc never passes test
                    # calculation converged, save the results
                    self.logger.pr('calculation label %d client %d CONVERGED' % (label, client_id))
                    results[label] = res
                else:
                    # calculation did not converge, we need to repeat it
                    self.logger.pr('calculation label %d client %d DID NOT CONVERGE' % (label, client_id))
                    rejects.append(-label)

                self.output_q.task_done()

            self.logger.pr('AtomsServer.get_results() rejects=%r' % rejects)
            self.logger.pr('AtomsServer.get_results() sorted(results.keys())=%r' % sorted(results.keys()))

            # collect all input task so we can see if anything is missing
            input = {}
            while self.input_q.unfinished_tasks:
                label, client_id, at = self.input_q.get()
                input[label] = (client_id, at)
                self.input_q.task_done()

            self.logger.pr('AtomsServer.get_results() sorted(input.keys())=%r' % sorted(input.keys()))

            # resubmit any failed calculations
            for label in rejects:
                client_id, at = input[label]
                self.logger.pr('Resubmiting calculation label %d client_id %d' % (label, client_id))
                self.put(at, client_id, label, force_restart=True)

            assert len(results) + len(rejects) == len(input)

            # if all calculations converged we are done
            if len(rejects) == 0:
                break
        else:
            raise RuntimeError('max_attempts (%d) exceeded without all calculations completing successfully' %
                               self.max_attempts)

        assert(len(results) == len(input))
        assert(len(rejects) == 0)

        results_atoms = []
        for (inp_label, label) in zip(sorted(input.keys()), sorted(results.keys())):

            assert inp_label == label
            client_id, inp_at = input[inp_label]
            res = results[label]

            if isinstance(res, Atoms):
                at = res
            else:
                (natoms, energy, force, virial) = res
                assert len(inp_at) == natoms

                at = inp_at.copy() # FIXME could possibly store results inplace, but need to think about sorting
                at.info['label'] = label
                at.info['energy'] = energy
                at.set_array('force', force)
                at.info['virial'] = virial

            # allow client to modify results (e.g. reverse sort order)
            at = self.clients[client_id].postprocess(at, label)

            results_atoms.append(at)

        self.logger.pr('AtomsServer processed %d results' % len(results))
        return results_atoms


class AtomsServerAsync(AtomsServerSync, socketserver.ThreadingMixIn):
    """
    Asynchronous (threaded) version of AtomsServer
    """

    def shutdown(self):
        self.shutdown_clients()
        return socketserver.TCPServer.shutdown(self)

    def shutdown_clients(self):
        self.logger.pr('shutting down all clients')
        wait_threads = []
        for client_id, client in enumerate(self.clients):
            if (client.process is not None and client.process.poll() is None and
                (client.wait_thread is None or not client.wait_thread.isAlive())):
                wait_threads.append(client.shutdown(block=False))
        # wait for them all to finish shutting down
        for wait_thread in wait_threads:
            if wait_thread is None or not wait_thread.isAlive():
                continue
            wait_thread.join()
        self.logger.pr('all client shutdowns complete')


AtomsServer = AtomsServerAsync # backwards compatibility

class Client(object):
    """
    Represents a single Client job

    Used by AtomsServer to start, restart and shutdown clients
    running on the Compute Nodes.
    """

    def __init__(self, client_id, exe, env=None, npj=1, ppn=1,
                 block=None, corner=None, shape=None,
                 jobname='socketcalc', rundir=None,
                 fmt='REFTRAJ', parmode=None, mpirun='mpirun',
                 mpirun_args=['-np'], logger=screen,
                 max_pos_diff=MAX_POS_DIFF,
                 max_cell_diff=MAX_CELL_DIFF):

        self.client_id = client_id
        self.process = None # handle for the runjob process
        self.log = None # stdout file
        self.wait_thread = None # used by shutdown(block=False)
        self.last_atoms = None # used to check if we can continue from previous task
        self.lock = threading.Lock() # avoid concurrancy issues

        if env is None:
            env = {}
        self.env = env # environment
        self.exe = exe # executable
        self.npj = npj # nodes per job
        self.ppn = ppn # processes per node

        self.block, self.corner, self.shape = block, corner, shape

        self.jobname = jobname
        self.fmt = fmt
        self.parmode = parmode
        self.mpirun = mpirun
        self.mpirun_args = mpirun_args
        self.logger = logger
        self.max_pos_diff = max_pos_diff
        self.max_cell_diff = max_cell_diff

        self.rundir = rundir or os.getcwd()
        self.subdir = os.path.join(self.rundir, '%s-%03d' % (jobname, self.client_id))
        if not os.path.exists(self.subdir):
            self.logger.pr('Making subdir %s' % self.subdir)
            os.mkdir(self.subdir)

    def extra_args(self, label=None):
        """
        Return list of additional command line arguments to be passed to client
        """
        args = [self.server.ip, str(self.server.port), str(self.client_id)]
        if label is not None:
            args.append(str(label))
        return args

    def start(self, label=None):
        """
        Start an individual client.

        Raises RuntimeError if this client is already running.
        """
        if self.process is not None:
            raise RuntimeError('client %d is already running' % client_id)

        runjob_args = []
        popen_args = {}
        if self.parmode == 'cobalt':
            # Convert env to "--envs KEY=value" arguments for runjob
            envargs = []
            for (k, v) in self.env.iteritems():
                envargs.extend(['--envs', '%s=%s' % (k, v) ])

            runjob_args += ['runjob', '--block', self.block]
            if self.corner is not None:
                runjob_args += ['--corner', self.corner]
            if self.shape is not None:
                runjob_args += ['--shape', self.shape]
            runjob_args += (['-n', str(self.npj*self.ppn), '-p', str(self.ppn)] + envargs +
                            ['--cwd', self.subdir, ':'])
        elif self.parmode == 'mpi':
            runjob_args += [self.mpirun]
            for mpirun_arg in self.mpirun_args:
                runjob_args += [mpirun_arg]
                if mpirun_arg in ['-n', '-np']:
                    runjob_args += [str(self.npj*self.ppn)]
            popen_args['cwd'] = self.subdir
            popen_args['env'] = os.environ # for mpi, let mpirun inherit environment of script
        else:
            popen_args['cwd'] = self.subdir
            popen_args['env'] = self.env
        runjob_args += [self.exe]
        runjob_args += self.extra_args(label)
        self.logger.pr('starting client %d args %r' % (self.client_id, runjob_args))
        self.log = open(os.path.join(self.rundir, '%s-%03d.output' % (self.jobname, self.client_id)), 'a')
        # send stdout and stderr to same file
        self.process = subprocess.Popen(runjob_args, stdout=self.log, stderr=self.log, **popen_args)



    def shutdown(self, block=True):
        """Request a client to shutdown.

        If block=True, does not return until shutdown is complete.  If
        block=False, waits for the client to shutdown in a new
        thread. Check self.waits_thread.isAlive() to see when shutdown
        has finished. (This function also returns a handle to the wait
        thread when block=False).
        """
        if self.process is None:
            self.logger.pr('client %d (requested to shutdown) has never been started' % self.client_id)
            return

        if self.process.poll() is not None:
            self.logger.pr('client %d is already shutdown' % self.client_id)
            return

        if (self.wait_thread is not None and self.wait_thread.isAlive()):
            raise RuntimeError('client %d is already in the process of shutting down' % self.client_id)

        input_q = self.server.input_qs[self.client_id]
        input_q.put((b'shutdown', self.fmt, -1, None))

        if block:
            self.wait_for_shutdown()
        else:
            self.wait_thread = threading.Thread(target=self.wait_for_shutdown)
            self.wait_thread.start()
            return self.wait_thread


    def wait_for_shutdown(self):
        """
        Block until a client has shutdown.

        Typically called automatically by shutdown() or
        start_or_restart().

        Shutdown should previously have been initiated by queuing a
        'shutdown' or 'restart' request. Waits CLIENT_TIMEOUT for
        graceful shutdown. If client is still alive, a SIGTERM signal
        is sent. If this has had no effect after a further
        CLIENT_TIMEOUT, then a SIGKILL is sent. Does not return until
        the SIGKILL has taken effect.

        This function also marks shutdown task as complete in
        servers's input_q for this client.
        """
        wait_thread = threading.Thread(target=self.process.wait)
        self.logger.pr('waiting for client %d to shutdown' % self.client_id)
        wait_thread.start()
        wait_thread.join(CLIENT_TIMEOUT)
        if wait_thread.isAlive():
            self.logger.pr('client %d did not shutdown gracefully in %d seconds - sending SIGTERM' %
                         (self.client_id, CLIENT_TIMEOUT))
            self.process.terminate()
            wait_thread.join(CLIENT_TIMEOUT)
            if wait_thread.isAlive():
                self.logger.pr('client %d did not respond to SIGTERM - sending SIGKILL' % self.client_id)
                self.process.kill()
                wait_thread.join() # no timeout for kill
            else:
                self.logger.pr('client %d responded to SIGTERM' % self.client_id)
        else:
            self.logger.pr('client %d shutdown within timeout' % self.client_id)
        self.logger.pr('client %d shutdown complete - exit code %r' % (self.client_id, self.process.poll()))
        self.log.close()

        self.process = None
        self.log = None
        self.server.input_qs[self.client_id].task_done()
        self.logger.pr('wait_for_shutdown done')


    def start_or_restart(self, at, label, restart=False):
        """
        Start or restart a client

        If restart=True, wait for previous client to shutdown first.
        Calls write_input_files() followed by start().
        """
        if restart:
            self.wait_for_shutdown()
        self.write_input_files(at, label)
        self.start(label)


    def preprocess(self, at, label, force_restart=False):
        """
        Prepare client for a calculation.

        Starts client if this is the first task for it, or schedules a
        restart if new configuration is not compatible with the last
        one submitted to the queue (see is_compatible() method).

        Many be extended in subclasses to e.g. sort the atoms by
        atomic number. If Atoms object needs to be changed, a copy
        should be returned rather than updating it inplace.

        Returns (at, first_time).
        """

        first_time = self.process is None
        restart_reqd = (not first_time and (force_restart or
                                            (not self.is_compatible(self.last_atoms, at, label))))

        # keep a copy of last config queued for this client.
        # acquire a lock in case multiple put() calls to the same client
        # occur concurrently.
        try:
            self.lock.acquire()
            self.last_atoms = at.copy()
        finally:
            self.lock.release()

        if restart_reqd:
            # put a shutdown command into the queue, ahead of this config.
            # once it gets completed, restart_client() will be called as below
            self.logger.pr('restart scheduled for client %d label %d' % (self.client_id, label))
            self.server.input_qs[self.client_id].put((b'restart', self.fmt, -1, None))
        if first_time:
            self.start_or_restart(at, label, restart=False)

        return at, self.fmt, first_time


    def postprocess(self, at, label):
        """
        Post-process results of calculation.

        May be overrriden in subclasses to e.g. reverse sort order
        applied in preprocess().
        """
        return at


    def is_compatible(self, old_at, new_at, label):
        """
        Check if new_at and old_at are compatible.

        Returns True if calculation can be continued, or False
        if client must be restarted before it can process new_at.
        """
        if old_at is None:
            return True

        return True


    def write_input_files(self, at, label):
        raise NotImplementedError('to be implemented in subclasses')



class QUIPClient(Client):
    """
    Subclass of Client for running QUIP calculations.

    Initial input files are written in extended XYZ format, and
    subsequent communication is via sockets, in either REFTRAJ
    or XYZ format.
    """

    def __init__(self, client_id, exe, env=None, npj=1, ppn=1,
                 block=None, corner=None, shape=None,
                 jobname='socketcalc', rundir=None,
                 fmt='REFTRAJ', parmode=None, mpirun='mpirun',
                 mpirun_args=['-np'], logger=screen,
                 max_pos_diff=MAX_POS_DIFF,
                 max_cell_diff=MAX_CELL_DIFF,
                 param_files=None):
        Client.__init__(self, client_id, exe, env, npj, ppn,
                        block, corner, shape, jobname, rundir, fmt, parmode,
                        mpirun, mpirun_args, logger, max_pos_diff,
                        max_cell_diff)
        self.param_files = param_files

    def write_input_files(self, at, label):
        write_xyz(os.path.join(self.subdir, 'atoms.%d.xyz' % self.client_id), at)

        # copy in parameter files
        if self.param_files is not None:
            for param_file in self.param_files:
                param_file_basename = os.path.basename(param_file)
                shutil.copyfile(param_file, os.path.join(self.subdir, param_file_basename))


_chdir_lock = threading.Lock()


class QMClient(Client):
    """
    Abstract subclass of Client for QM calculations
    """

    def is_compatible(self, old_at, new_at, label):
        # first time, anything goes
        if old_at is None:
            return True

        if not Client.is_compatible(self, old_at, new_at, label):
            return False

        if len(old_at) != len(new_at):
            self.logger.pr('is_compatible() on client %d label %d got number of atoms mismatch: %d != %d' % (self.client_id,
                                                                                                           label,
                                                                                                           len(old_at),
                                                                                                           len(new_at)))
            return False # number of atoms must match

        if abs(old_at.cell - new_at.cell).max() > self.max_cell_diff:
            self.logger.pr('is_compatible() on client %d label %d got cell mismatch: %r != %r' % (self.client_id,
                                                                                                   label,
                                                                                                   old_at.cell,
                                                                                                   new_at.cell))
            return False # cells must match

        # RMS difference in positions must be less than max_pos_diff
        old_p = old_at.get_positions()
        new_p = new_at.get_positions()

        old_z = old_at.get_chemical_symbols()
        new_z = new_at.get_chemical_symbols()

        if 'index' in old_at.arrays:
            old_index = old_at.get_array('index')
            new_index = new_at.get_array('index')

            # if termination exists, undo ordering differences due to cluster hopping
            if ('termindex_%d' % self.client_id) in old_at.arrays:
                old_termindex = old_at.get_array('termindex_%d' % self.client_id)
                new_termindex = new_at.get_array('termindex_%d' % self.client_id)

                a1s = sorted([(old_index[i], old_z[i], list(old_p[i]))
                              for i in range(len(old_at)) if old_termindex == 0])

                a2s = sorted([(new_index[i], new_z[i], list(new_p[i]))
                              for i in range(len(new_at)) if new_termindex == 0])
            else:
                a1s = sorted([(old_index[i], old_z[i], list(old_p[i])) for i in range(len(old_at)) ])
                a2s = sorted([(new_index[i], new_z[i], list(new_p[i])) for i in range(len(new_at)) ])

            old_p = np.r_[[p for (i, z, p) in a1s]]
            new_p = np.r_[[p for (i, z, p) in a2s]]

            old_z = np.r_[[z for (i, z, p) in a1s]]
            new_z = np.r_[[z for (i, z, p) in a2s]]

        if not np.all(old_z == new_z):
            self.logger.pr('is_compatible() on client %d label %d got atomic number mismatch: %r != %r' % (self.client_id,
                                                                                                         label,
                                                                                                         old_z, new_z))
            return False # atomic numbers must match

        # undo jumps across PBC - approach is that of QUIP's undo_pbc_jumps() routine
        old_g = np.linalg.inv(old_at.cell.T).T
        d = new_p.T - old_p.T - (np.dot(old_at.cell, np.floor(np.dot(old_g, (new_p - old_p).T)+0.5)))
        rms_diff = np.sqrt((d**2).mean())
        self.logger.pr('is_compatible() on client %d label %d got RMS position difference %.3f' % (self.client_id, label, rms_diff))

        if rms_diff > self.max_pos_diff:
            self.logger.pr('is_compatible() on client %d label %d got RMS position difference %.3f > max_pos_diff=%.3f' %
                                  (self.client_id, label, rms_diff, self.max_pos_diff))
            return False

        return True


class VaspClient(QMClient):
    """
    Subclass of Client for running VASP calculations.

    Initial input files are written in POSCAR, INCAR, POTCAR and KPOINTS
    formats, and subsequent communicatin is via sockets in REFTRAJ format.
    """

    def __init__(self, client_id, exe, env=None, npj=1, ppn=1,
                 block=None, corner=None, shape=None,
                 jobname='socketcalc', rundir=None,
                 fmt='REFTRAJ', parmode=None, mpirun='mpirun',
                 mpirun_args=['-np'], logger=screen,
                 max_pos_diff=MAX_POS_DIFF,
                 max_cell_diff=MAX_CELL_DIFF,
                 **vasp_args):
        Client.__init__(self, client_id, exe, env, npj, ppn,
                        block, corner, shape, jobname, rundir,
                        fmt, parmode, mpirun, mpirun_args, logger,
                        max_pos_diff, max_cell_diff)
        if 'ibrion' not in vasp_args:
            self.logger.pr('No ibrion key in vasp_args, setting ibrion=13')
            vasp_args['ibrion'] = 13
        if 'nsw' not in vasp_args:
            self.logger.pr('No nsw key in vasp_args, setting nsw=1000000')
            vasp_args['nsw'] = 1000000
        self.vasp_args = vasp_args


    def preprocess(self, at, label, force_restart=False):
        self.logger.pr('vasp client %d preprocessing atoms label %d' % (self.client_id, label))

        # make a copy and then sort atoms in the same way that vasp
        # calculator will when it writes POSCAR. We use a new
        # calculator and store the sort order in the Atoms so it can
        # be reversed when results are ready.
        vasp = Vasp(**self.vasp_args)
        vasp.initialize(at)
        at = at.copy()
        order = np.array(range(len(at)))
        at.set_array('vasp_sort_order', order)
        at = at[vasp.resort]

        # finally, call the parent method
        return Client.preprocess(self, at, label, force_restart)


    def postprocess(self, at, label):
        self.logger.pr('vasp client %d postprocessing atoms label %d' % (self.client_id, label))
        # call the parent method first
        at = Client.postprocess(self, at, label)
        # restore original atom ordering
        at = at[at.arrays['vasp_sort_order'].tolist()]
        return at


    def write_input_files(self, at, label):
        global _chdir_lock
        # For LOTF Simulations active number of quantum 
        # atoms vary and must wait to this stage in order for
        # magnetic moments to be set properly. If magnetic moments
        # not set defaults to 0.
        self.vasp_args['magmom'] = at.get_initial_magnetic_moments()
        vasp = Vasp(**self.vasp_args)
        vasp.initialize(at)
        # chdir not thread safe, so acquire global lock before using it
        orig_dir = os.getcwd()
        try:
            _chdir_lock.acquire()
            os.chdir(self.subdir)
            if os.path.exists('OUTCAR'):
                n = 1
                while os.path.exists('OUTCAR.%d' % n):
                    n += 1
                shutil.copyfile('OUTCAR', 'OUTCAR.%d' % n)
                shutil.copyfile('POSCAR', 'POSCAR.%d' % n)
            write_vasp('POSCAR', vasp.atoms_sorted,
                       symbol_count=vasp.symbol_count,
                       vasp5='5' in self.exe)
            vasp.write_incar(at)
            vasp.write_potcar()
            vasp.write_kpoints()
        finally:
            os.chdir(orig_dir)
            _chdir_lock.release()


class CastepClient(QMClient):
    """
    Subclass of Client for running CASTEP calculations.

    Initial input files are written in .cell and .param
    formats, and subsequent communication is via sockets in REFTRAJ format.
    """
    def __init__(self, client_id, exe, env=None, npj=1, ppn=1,
                 block=None, corner=None, shape=None,
                 jobname='socketcalc', rundir=None,
                 fmt='REFTRAJ', parmode=None, mpirun='mpirun',
                 mpirun_args=['-np'], logger=screen,
                 max_pos_diff=MAX_POS_DIFF_CASTEP,
                 max_cell_diff=MAX_CELL_DIFF_CASTEP,
                 **castep_args):
        Client.__init__(self, client_id, exe, env, npj, ppn,
                        block, corner, shape, jobname, rundir,
                        fmt, parmode, mpirun, mpirun_args, logger,
                        max_pos_diff, max_cell_diff)
        if 'task' not in castep_args:
            self.logger.pr('No task key in castep_args, setting task=MD')
            castep_args['task'] = 'MD'
        if 'md_ensemble' not in castep_args:
            self.logger.pr('No md_ensemble key in castep_args, setting md_ensemble=SOC')
            castep_args['md_ensemble'] = 'SOC'
        if 'md_num_iter' not in castep_args:
            self.logger.pr('No md_num_iter key in castep_args, setting md_num_iter=1000000')
            castep_args['md_num_iter'] = 1000000
        castep_args['_rename_existing_dir'] = False
        self.castep_args = castep_args
        self.logger.pr('constructing Castep instance with args %r' % castep_args)
        self.castep = Castep(directory=self.subdir, **castep_args)

        self._orig_devel_code = ''
        if self.castep.param.devel_code.value is not None:
            self._orig_devel_code = self.castep.param.devel_code.value.strip()+'\n'

    def preprocess(self, at, label, force_restart=False):
        self.logger.pr('Castep client %d preprocessing atoms label %d' % (self.client_id, label))

        # make a copy and then sort atoms by atomic number
        # in the same way that Castep will internally. We store the sort
        # order in the Atoms so it can be reversed when results are ready.
        at = at.copy()
        order = np.array(range(len(at)))
        at.set_array('castep_sort_order', order)
        resort = order[np.argsort(at.get_atomic_numbers())]
        #print 'resort = ', resort
        #print at.get_scaled_positions()[resort[0]]
        at = at[resort]
        #print at.get_scaled_positions()[0]
        #print 'castep_sort_order', at.get_array('castep_sort_order')

        # finally, call the parent method (potentially writing input files)
        return Client.preprocess(self, at, label, force_restart)


    def postprocess(self, at, label):
        self.logger.pr('Castep client %d postprocessing atoms label %d' % (self.client_id, label))
        # call the parent method first
        at = Client.postprocess(self, at, label)
        # restore original atom ordering
        at = at[at.arrays['castep_sort_order'].tolist()]
        return at

    def write_input_files(self, at, label):
        global _chdir_lock

        devel_code = self._orig_devel_code
        devel_code += ('SOCKET_IP=%s\nSOCKET_PORT=%d\nSOCKET_CLIENT_ID=%d\nSOCKET_LABEL=%d' % \
                        (self.server.ip, self.server.port, self.client_id, label))
        self.castep.param.devel_code = devel_code

        # chdir not thread safe, so acquire global lock before using it
        orig_dir = os.getcwd()
        try:
            _chdir_lock.acquire()
            os.chdir(self.subdir)
            cellf = open('castep.cell', 'w')
            write_castep_cell(cellf, at, castep_cell=self.castep.cell)
            cellf.close()
            write_param('castep.param', self.castep.param, force_write=True)

        finally:
            os.chdir(orig_dir)
            _chdir_lock.release()

    def extra_args(self, label=None):
        return ['castep']


class SocketCalculator(Calculator):
    """
    ASE-compatible calculator which communicates with remote
    force engines via sockets using a (synchronous) AtomsServer.
    """

    implemented_properties = ['energy', 'forces', 'stress']
    default_parameters = {}
    name = 'SocketCalculator'

    def __init__(self, client, ip=None, atoms=None, port=0, logger=screen, bgq=False):
        Calculator.__init__(self)

        self.client = client
        if ip is None:
            ip = '127.0.0.1' # default to localhost
        self.logger = logger
        self.bgq=bgq
        self.server = AtomsServerSync((ip, port), AtomsRequestHandler,
                                      [self.client], logger=self.logger,
                                      bgq=self.bgq)
        self._label = 1
        self.atoms = atoms

    def calculate(self, atoms, properties, system_changes):
        Calculator.calculate(self, atoms, properties, system_changes)
        if system_changes: # if anything at all changed (could be made more fine-grained)
            self.logger.pr('calculation triggered with properties={0}, system_changes={1}'.format(properties,
                                                                                                  system_changes))
            self.server.put(atoms, 0, self._label)
            if self._label != 1:
                # send atoms over socket, unless first time
                self.logger.pr('socket calculator sending Atoms label={0}'.format(self._label))
                self.server.handle_request()
            # wait for results to be ready
            self.logger.pr('socket calculator waiting for results label={0}'.format(self._label))
            self.server.handle_request()

            self._label += 1
            [results] = self.server.get_results()

            # we always compute energy, forces and stresses, regardless of what was requested
            stress = -(results.info['virial']/results.get_volume())
            self.results = {'energy': results.info['energy'],
                            'forces': results.arrays['force'],
                            'stress': full_3x3_to_Voigt_6_stress(stress)}
        else:
            self.logger.pr('calculation avoided with properties={0}, system_changes={1}'.format(properties,
                                                                                                system_changes))

    def shutdown(self):
        self.server.shutdown()




# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

"""
Contact mechanics module.

This module contains functions creating real and reciprocal space Green's
function. Continuous real space Green's function can be converted to their
discrete reciprocal space representation by `real_to_reciprocal_space`.
Typically, all *nonperiodic* kernel return real space Green's functions,
all *periodic* kernel reciprocal space ones.
"""

from __future__ import division

from math import isnan, pi, sqrt

import numpy as np

###

def square_pressure__nonperiodic(x, y, a=0.5, b=0.5):
    """
    Real-space representation of Green's function for the normal displacements
    of a non-periodic linear elastic half-space with contact modulus 2 and
    Poisson number 1/2 in response to a uniform pressure applied to a
    rectangular region.
    See: K.L. Johnson, Contact Mechanics, p. 54

    Parameters
    ----------
    x : array_like
        x-coordinates.
    y : array_like
        y-coordinates.
    a, b : float
        Size of rectangle. Rectangle has corners at -a,-b and a,b, i.e.
        rectangle has extend of 2*a x 2*b. Solution converges to
        Boussinesq-Cerrutti form for a->0 and b->0.

    Returns
    -------
    gf : array
        Green's function for displacements at coordinates given by x,y as a
        function of pressure applied to the rectangular region.
    """

    return ( (x+a)*np.log( ( (y+b)+np.sqrt((y+b)*(y+b)+(x+a)*(x+a)) )/
                           ( (y-b)+np.sqrt((y-b)*(y-b)+(x+a)*(x+a)) ) )+
             (y+b)*np.log( ( (x+a)+np.sqrt((y+b)*(y+b)+(x+a)*(x+a)) ) /
                           ( (x-a)+np.sqrt((y+b)*(y+b)+(x-a)*(x-a)) ) )+
             (x-a)*np.log( ( (y-b)+np.sqrt((y-b)*(y-b)+(x-a)*(x-a)) ) /
                           ( (y+b)+np.sqrt((y+b)*(y+b)+(x-a)*(x-a)) ) )+
             (y-b)*np.log( ( (x-a)+np.sqrt((y-b)*(y-b)+(x-a)*(x-a)) ) /
                           ( (x+a)+np.sqrt((y-b)*(y-b)+(x+a)*(x+a)) ) ) )/(2*pi);


def point_traction__nonperiodic(quantities, x, y, z, G=1.0, poisson=0.5):
    """
    Real-space representation of Green's function for the displacement and
    stress in the bulk of a non-periodic linear elastic half-space in response
    to a concentrated surface force. This is the Boussinesq-Cerrutti solution.
    See: K.L. Johnson, Contact Mechanics, p. 51 and p. 69
    Sign convention is as in Johnson!

    Parameters
    ----------
    quantities : str
        Each character in this string defines a return quantity. They are
        returned in a tuple of the same order. Possible quantities are
            'x' : Displacement for a concentrated surface traction in
                  x-direction.
            'X' : Stress for a concentrated surface traction in x-direction.
            'Y' : Stress for a concentrated surface traction in y-direction.
            'Z' : Stress for a concentrated surface pressure
                  (i.e. "traction" in z-direction).
    x : array_like
        x-coordinates.
    y : array_like
        y-coordinates.
    z : array_like
        z-coordinates. Into the solid is positive.
    G : float
        Shear modulus.
    poisson : float
        Poisson number.

    Returns
    -------
    sxx : array
        Green's function xx-component of the stress tensor at coordinates given
        by x,y,z as a function of pressure applied at point 0,0.
    syy : array
        yy-component of the stress tensor.
    szz : array
        zz-component of the stress tensor.
    syz : array
        yz-component of the stress tensor.
    sxz : array
        xz-component of the stress tensor.
    sxy : array
        xy-component of the stress tensor.
    """

    r_sq = x**2 + y**2
    r_sq = np.array(r_sq, dtype=float)
    rho = np.sqrt(r_sq + z**2)

    r_sq[r_sq <= 0.0] = 1e-9
    rho[rho <= 0.0] = 1e-9

    retvals = []
    for q in quantities:
        if q == 'x':
            ux = (1/rho+x**2/rho**3+(1-2*poisson)*(1/(rho+z)-x**2/(rho*(rho+z)**2)))/(4*pi*G)
            uy = (x*y/rho**3-(1-2*poisson)*x*y/(rho*(rho+z)**2))/(4*pi*G)
            uz = (x*z/rho**3+(1-2*poisson)*x/(rho*(rho+z)))/(4*pi*G)
            retvals += [np.array([ux, uy, uz])]
        elif q == 'y':
            raise NotImplementedError()
        elif q == 'z':
            ux = (x*z/rho**3-(1-2*poisson)*x/(rho*(rho+z)))/(4*pi*G)
            uy = (y*z/rho**3-(1-2*poisson)*y/(rho*(rho+z)))/(4*pi*G)
            uz = (z**2/rho**3+2*(1-poisson)/rho)/(4*pi*G)
            retvals += [np.array([ux, uy, uz])]
        elif q == 'X':
            sxx = ( -3*x**3/rho**5 + (1-2*poisson)*(x/rho**3 - 3*x/(rho*(rho+z)**2) + x**3/(rho**3*(rho+z)**2) + 2*x**3/(rho**2*(rho+z)**3)) )/(2*pi)
            syy = ( -3*x*y**2/rho**5 + (1-2*poisson)*(x/rho**3 - x/(rho*(rho+z)**2) + x*y**2/(rho**3*(rho+z)**2) + 2*x*y**2/(rho**2*(rho+z)**3)) )/(2*pi)
            szz = ( -3*x*z**2/rho**5 )/(2*pi)
            sxy = ( -3*x**2*y/rho**5 + (1-2*poisson)*(-y/(rho*(rho+z)**2) + x**2*y/(rho**3*(rho+z)**2) + 2*x**2*y/(rho**2*(rho+z)**3)) )/(2*pi)
            sxz = ( -3*x**2*z/rho**5 )/(2*pi)
            syz = ( -3*x*y*z/rho**5 )/(2*pi)
            retvals += [np.array([sxx, syy, szz, syz, sxz, sxy])]
        elif q == 'Y':
            raise NotImplementedError()
        elif q == 'Z':
            sxx = ( (1-2*poisson)/r_sq * ((1 - z/rho) * (x**2 - y**2)/r_sq + z*y**2/rho**3) - 3*z*x**2/rho**5 )/(2*pi)
            syy = ( (1-2*poisson)/r_sq * ((1 - z/rho) * (y**2 - x**2)/r_sq + z*x**2/rho**3) - 3*z*y**2/rho**5 )/(2*pi)
            szz = -3*z**3/(2*pi*rho**5)
            # Note: Johnson is lacking a factor of 2 here!!!
            sxy = ( (1-2*poisson)/r_sq * (2*(1 - z/rho) * x*y/r_sq - x*y*z/rho**3) - 3*x*y*z/rho**5 )/(2*pi)
            sxz = -3*x*z**2/(2*pi*rho**5)
            syz = -3*y*z**2/(2*pi*rho**5)
            retvals += [np.array([sxx, syy, szz, syz, sxz, sxy])]
        else:
            raise ValueError("Unknown quantity '{0}' requested.".format(q))

    if len(quantities) == 1:
        return retvals[0]
    else:
        return retvals


def real_to_reciprocal_space(nx, ny=None, gf=square_pressure__nonperiodic,
                             coordinates=False):
    """
    Return the reciprocal space representation of a real-space Green's function
    on an FFT grid.
    Note: If the Green's function is for the non-periodic (free-bounday)
    problem, then only a section of 1/2 nx by 1/2 ny of the grid can have
    non-zero pressure. The other region is a padding region. See R.W. Hockney,
    Methods in Computational Physics, Vol. 9, pp. 135-211 (1970) for an
    application of this method to the electrostatic problem.

    Parameters
    ----------
    nx : int
        Number of grid points in x-direction.
    ny : int
        Number of grid points in y-direction.
    gf : function
        Function returning the real-space Green's function.
    coordinates : bool
        If True, return grid coordinates.

    Returns
    -------
    G : array
        Reciprocal space representation of the stiffness on an nx,ny grid.
    x : array
        x-coordinates of grid.
    y : array
        y-coordinates of grid.
    """

    if ny is None:
        nx, ny = nx

    x = np.arange(nx)
    x = np.where(x <= nx//2, x, x-nx)
    x.shape = (-1,1)
    y = np.arange(ny)
    y = np.where(y <= ny//2, y, y-ny)
    y.shape = (1,-1)

    G = gf(x, y)
    if isinstance(G, tuple) or isinstance(G, list):
        r = [ np.fft.fft2(np.real(_G)) for _G in G ]
    else:
        r = np.fft.fft2(np.real(G))

    if coordinates:
        return r, x, y
    else:
        return r


def point_displacement__periodic(nx, ny=None, phi0=None, size=None):
    """
    Return reciprocal space stiffness coefficients (i.e. inverse of the Green's
    function) for a periodic system with contact modulus 2 and Poisson number
    1/2. This gives force as a function of displacement.

    Parameters
    ----------
    nx : int
        Number of grid points in x-direction.
    ny : int
        Number of grid points in y-direction.
    phi0 : float
        Stiffness at q=0 (gamma point).
    size : tuple
        System size. Assumed to be nx,ny if parameter is omitted.

    Returns
    -------
    phi : array
        Reciprocal space representation of the stiffness on an nx,ny grid.
    """

    if ny is None:
        nx, ny = nx

    qx = np.arange(nx, dtype=np.float64)
    qx = np.where(qx <= nx//2, 2*pi*qx/sx, 2*pi*(nx-qx)/sx)
    qy = np.arange(ny, dtype=np.float64)
    qy = np.where(qy <= ny//2, 2*pi*qy/sy, 2*pi*(ny-qy)/sy)
    phi  = np.sqrt( (qx*qx).reshape(-1, 1) + (qy*qy).reshape(1, -1) )
    if phi0 is None:
        phi[0, 0] = (phi[1, 0].real + phi[0, 1].real)/2
    else:
        phi[0, 0] = phi0

    return phi


def min_ccg(h_r, gf_q, u_r=None, pentol=1e-6, maxiter=100000,
            logger=None):
    """
    Use a constrained conjugate gradient optimization to find the equilibrium
    configuration deflection of an elastic manifold. The conjugate gradient
    iteration is reset using the steepest descent direction whenever the contact
    area changes.
    Method is described in I.A. Polonsky, L.M. Keer, Wear 231, 206 (1999)

    Parameters
    ----------
    h_r : array_like
        Height profile of the rigid counterbody.
    gf_q : array_like
        Green's function (in reciprocal space).
    u_r : array
        Array used for initial displacements. A new array is created if omitted.
    pentol : float
        Maximum penetration of contacting regions required for convergence.
    maxiter : float
        Maximum number of iterations.

    Returns
    -------
    u : array
        2d-array of displacements.
    p : array
        2d-array of pressure.
    """

    # Note: Suffix _r deontes real-space _q reciprocal space 2d-arrays

    if logger is not None:
        logger.pr('maxiter = {0}'.format(maxiter))
        logger.pr('pentol = {0}'.format(pentol))

    if u_r is None:
        u_r = np.zeros_like(h_r)

    u_r[:, :] = np.where(u_r > h_r, h_r, u_r)

    # Compute forces
    p_r = -np.fft.ifft2(np.fft.fft2(u_r)/gf_q).real

    # iteration
    delta = 0
    delta_str = 'reset'
    G_old = 1.0
    for it in range(1, maxiter+1):
        # Reset contact area
        c_r = p_r > 0.0

        # Compute total contact area (area with repulsive force)
        A = np.sum(c_r)

        # Compute G = sum(g*g) (over contact area only)
        g_r = h_r-u_r
        G = np.sum(c_r*g_r*g_r)

        # t = (g + delta*(G/G_old)*t) inside contact area and 0 outside
        if delta > 0:
            t_r = c_r*(g_r + delta*(G/G_old)*t_r)
        else:
            t_r = c_r*g_r

        # Compute elastic displacement that belong to t_r
        # (Note: r_r is negative of Polonsky, Kerr's r)
        r_r = -np.fft.ifft2(gf_q*np.fft.fft2(t_r)).real

        # Note: Sign reversed from Polonsky, Keer because this r_r is negative
        # of theirs.
        tau = 0.0
        if A > 0:
            # tau = -sum(g*t)/sum(r*t) where sum is only over contact region
            x = -np.sum(c_r*r_r*t_r)
            if x > 0.0:
                tau = np.sum(c_r*g_r*t_r)/x
            else:
                G = 0.0

        # Save forces for later
        fold_r = p_r.copy()

        p_r -= tau*c_r*t_r

        # Find area with negative forces and negative gap
        # (i.e. penetration of the two surfaces)
        nc_r = np.logical_and(p_r <= 0.0, g_r < 0.0)

        # Set all negative forces to zero
        p_r *= p_r > 0.0

        if np.sum(nc_r) > 0:
            # nc_r contains area that just jumped into contact. Update their
            # forces.
            p_r -= tau*nc_r*g_r
        
            delta = 0
            delta_str = 'sd'
        else:
            delta = 1
            delta_str = 'cg'

        # Compute new displacements from updated forces
        u_r = -np.fft.ifft2(gf_q*np.fft.fft2(p_r)).real
       
        # Store G for next step
        G_old = G
        
        # Compute root-mean square penetration, max penetration and max force
        # difference between the steps
        if A > 0:
            rms_pen = sqrt(G/A)
        else:
            rms_pen = sqrt(G)
        max_pen = max(0.0, np.max(c_r*(u_r-h_r)))

        # Elastic energy would be
        # e_el = -0.5*np.sum(p_r*u_r)

        if rms_pen < pentol and max_pen < pentol:
            if logger is not None:
                logger.st(['status', 'it', 'A', 'tau', 'rms_pen', 'max_pen'],
                          ['CONVERGED', it, A, tau, rms_pen, max_pen],
                          force_print=True)
            return u_r, p_r

        if logger is not None:
            logger.st(['status', 'it', 'A', 'tau', 'rms_pen', 'max_pen'],
                      [delta_str, it, A, tau, rms_pen, max_pen])

        if isnan(G) or isnan(rms_pen):
            raise RuntimeError('nan encountered.')

    raise RuntimeError('Maximum number of iterations ({0}) exceeded.' \
                           .format(maxiter))

